// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 * <p>
 * Represents an attribute such as <code>class="empty"</code> or
 * <code>xlink:href="http://www.example.com"</code>.
 * </p>
 * 
 * <p>
 * Attributes that declare namespaces such as
 * <code>xmlns="http://www.w3.org/TR/1999/xhtml"</code>
 * or <code>xmlns:xlink="http://www.w3.org/TR/1999/xlink"</code>
 * are stored separately on the elements where they
 * appear. They are never represented as <code>Attribute</code>
 * objects.
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class Attribute extends Node {
    
    private String localName;
    private String prefix;
    private String URI;
    private String value = "";
    private Type   type;

    /**
     * <p>
     * Creates a new attribute in no namespace with the
     * specified name and value and undeclared type.
     * </p>
     * 
     * @param localName the unprefixed attribute name
     * @param value the attribute value
     * 
     * @throws IllegalNameException if the local name is not 
     *                              a namespace well-formed non-colonized name
     * @throws IllegalDataException if the value contains characters which are 
     *   not legal in XML such as vertical tab or a null. Characters such as 
     *   " and &amp; are legal, but will be automatically 
     *   escaped when the attribute is serialized.
     */
    public Attribute(String localName, String value) {
        this(localName, "", value, Type.UNDECLARED);
    }

    /**
     * Creates a new attribute in no namespace with the
     * specified name, value, and type.
     * 
     * @param localName the unprefixed attribute name
     * @param value the attribute value
     * @param type the attribute type
     * 
     * @throws IllegalNameException if the local name is not a namespace well-formed non-colonized name
     * @throws IllegalDataException if the value contains characters which are not legal in
     * XML such as vertical tab or a null. Note that characters such as " and &amp; are legal,
     * but will be automatically escaped when the attribute is serialized.
     */
    public Attribute(String localName, String value, Type type) {
        this(localName, "", value, type);
    }

    /**
     * Creates a new attribute in the specified namespace with the
     * specified name and value and undeclared type.
     * 
     * @param name the prefixed attribute name
     * @param URI the namespace URI
     * @param value the attribute value
     * 
     * @throws IllegalNameException  if the name is not a namespace well-formed prefixed name
     * @throws IllegalDataException if the value contains characters which are not legal in
     * XML such as vertical tab or a null. Note that characters such as " and &amp; are legal,
     * but will be automatically escaped when the attribute is serialized.
     */
    public Attribute(String name, String URI, String value) {
        this(name, URI, value, Type.UNDECLARED);
    }

    /**
     * Creates a new attribute in the specified namespace with the
     * specified name and value and undeclared type.
     * 
     * @param name  the prefixed attribute name
     * @param URI the namespace URI
     * @param value the attribute value
     * @param type the attribute type
     * 
     * @throws IllegalNameException if the name is not a namespace well-formed prefixed name
     * @throws IllegalDataException if the value contains characters which are not legal in
     * XML such as vertical tab or a null. Note that characters such as " and &amp; are legal,
     * but will be automatically escaped when the attribute is serialized.
     */
    public Attribute(String name, String URI, String value, Type type) {

        // Move this into setName????
        prefix = "";
        String localName = name;
        if (name.indexOf(':') >= 0) {
            prefix = name.substring(0, name.indexOf(':'));   
            localName = name.substring(name.indexOf(':') + 1);   
        }

        setNamespace(prefix, URI);
        setLocalName(localName);
        setValue(value);
        setType(type);
    }

    /**
     * Creates a copy of the specified attribute.
     * 
     * @param attribute  the attribute to copy
     * 
     */
    public Attribute(Attribute attribute) {
        this(attribute.getQualifiedName(), attribute.URI, attribute.value, attribute.type);
    }

    /**
     * Returns the DTD type of this attribute. 
     * If this attribute does not have a type, then
     * Type.UNDECLARED is returned.
     * 
     * @return the DTD type of this attribute
     */
    public final Type getType() {
        return type;
    }

    /**
     * Sets the type of this attribute to one of the ten
     * DTD types or Type.UNDECLARED. 
     * 
     * @param type the DTD type of this attribute
     */
    public final void setType(Type type) {
        checkType(type);
        this.type = type;
    }

    /**
     * This method is called before the specified type is assigned to this attribute.
     * By default it does nothing. However,
     * subclasses can override it and prevent certain
     * types from being assigned to the attribute 
     * by throwing an XMLException.
     * 
     * @param type the DTD type of this attribute
     * 
     * @throws XMLException if the subclass rejects the requested type.
     */
    protected void checkType(Type type) {};

    /**
     * Returns the attribute value. If the attribute was
     * originally created by a parser then, it will have been
     * normalized according to its type.
     * However, attributes created in memory are not normalized.
     * 
     * @see nu.xom.Node#getValue()
     * 
     * @return the value of the attribute
     */
    public final String getValue() {
        return value;
    }

    /**
     * Sets the attribute's value to the speciifed string,
     * replacing any previous value. The value is not normalized
     * automatically.
     * 
     * @param value the value assigned to the attribute
     * @throws IllegalDataException if the value contains characters which are not legal in
     * XML such as vertical tab or a null. Note that characters such as " and &amp; are legal,
     * but will be automatically escaped when the attribute is serialized.
     */
    public final void setValue(String value) {
        Verifier.checkCharacterData(value);
        checkValue(value);
        this.value = value;
    }

    /**
     * Subclasses can override this method to check the value for 
     * an attribute before it's created or its value is changed.
     * 
     * @param value
     * 
     * @throws XMLException if value does not satisfy the local constraints
     * 
     */
    protected void checkValue(String value) {}


    /**
     * Returns the local name of this attribute,
     * not including the prefix.
     * 
     * @return the attribute's qualfied name
     */
    public final String getLocalName() {
        return localName;
    }

    /**
     * Sets the local name of the attribute.
     * 
     * @param localName the new local name
     * 
     * @throws IllegalNameException  if localName is not a legal, non-colonized name
     * 
     */
    public final void setLocalName(String localName) {
        Verifier.checkAttributeLocalName(localName);
        checkLocalName(localName);
        this.localName = localName;
    }

    /**
     * Returns the qualified name of this attribute,
     * including the prefix if this attribute is in a namespace.
     * 
     * @return the attribute's qualfied name
     */
    public final String getQualifiedName() {
        if (prefix.equals("")) return localName;
        else return prefix + ":" + localName;
    }
    
    
    /**
     * <p>
     * Returns the namespace URI of this attribute,
     * or the empty string if this attribute is not in a namespace.
     * </p>
     * 
     * Open question: should this be called just getNamespace or getURI????
     * 
     * @return the attribute's namespace URI
     */ 
    public final String getNamespaceURI() {
        return URI;
    }

    /**
     * Returns the prefix of this attribute,
     * or the empty string if this attribute is not in a namespace.
     * 
     * @return the attribute's prefix
     */
    public final String getNamespacePrefix() {
        return prefix;
    }

    /**
     * This method is called before the specified 
     * local name is assigned to this attribute.
     * By default it does nothing. However,
     * subclasses can override it and prevent certain
     * local names from being assigned to the attribute 
     * by throwing an XMLException.
     * 
     * @param localName the proposed local name of the attribute
     * @throws XMLException if the subclass rejects the name.
     */
    protected void checkLocalName(String localName) {}


    /**
     * Sets the attribute's namespace URI and prefix.
     * because attributes must be prefixed in order to have a namespace URI 
     * (and vice versa) this must be done simultaneously.
     * 
     * @param prefix the new namespace prefix
     * @param URI the new namespace URI
     * 
     * @throws NamespaceException if
     *  <ul>
     *      <li>The prefix is xmlns</li>
     *      <li>The prefix is xml and the namespace URI is not http://www.w3.org/XML/1998/namespace</li>
     *      <li>The prefix is null or the thempty string.</li>
     *      <li>The URI is null or the the empty string.</li>
     * </ul>
     */
    public final void setNamespace(String prefix, String URI) {
        
        if (URI == null) URI = "";
        if (prefix == null) prefix = "";
        
        if (prefix.equals("xmlns")) {
            throw new NamespaceException("Attribute objects are not used to represent namespace declarations"); 
        }
        else if (prefix.equals("xml") && !(URI.equals("http://www.w3.org/XML/1998/namespace"))) {
            throw new NamespaceException("Wrong namespace URI for xml prefix: " + URI); 
        }
        else if ("".equals(prefix)) {
            if ("".equals(URI)) {
                this.prefix = "";
                this.URI = "";
                return; 
            }
            else {
                throw new NamespaceException(
                  "Attributes cannot be in a default namespace.");
            }
        }
        else if ("".equals(URI)) {
            throw new NamespaceException(
             "Attribute prefixes must be declared.");
        }
        
        Verifier.checkNamespaceURI(URI);
        Verifier.checkNCName(prefix);
        checkNamespace(prefix, URI);
        
        this.URI = URI;
        this.prefix = prefix;
    }

    /**
     * This method is called before the specified 
     * local name is assigned to this attribute.
     * By default it does nothing. However,
     * subclasses can override it and prevent certain
     * namespace URIs from being assigned to the attribute 
     * by throwing an XMLException.
     * 
     * @param prefix the proposed namespace prefix for this attribute
     * @param URI the proposed namespace URI for this attribute
     * 
     * @throws XMLException if the subclass rejects the name.
     */
    protected void checkNamespace(String prefix, String URI) {}


    /**
     * Returns a string representation of the attribute suitable for 
     * debugging and diagnosis. However, this is not necessarily 
     * a well-formed XML attribute.
     * 
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        return "[" + getClass().getName() + ": " 
         + getQualifiedName() + "=\"" + getValue() + "\"]";
    }

    /**
     * Returns a string representation of the attribute 
     * that is a well-formed XML attribute. 
     * 
     * @see java.lang.Object#toString()
     */
    public final String getStringForm() {
        return getQualifiedName() + "=\"" 
         + escapeText(value) + "\"";    
    }

    private static String escapeText(String s) {
        
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < s.length(); i++) {
            result.append(escapeChar(s.charAt(i)));
        }
        return result.toString();
    }
    
    private static String escapeChar(char c) {
        
        switch (c) {
            case '<':
                return "&lt;";
            case '>':
                return "&gt;";
            case '&':
                return "&amp;";
            case '"':
                return "&quot;";
            case '\'':
                return "&apos;";
            default: 
                return String.valueOf(c);   
        }
        
    }
    


    /**
     * <p>
     * Uses the type-safe enumeration 
     * design pattern to represent attribute types,
     * as used in XML DTDs.
     * </p>
     * 
     * <p>
     *   XOM enforces well-formedness, but it does not enforce validity.
     *   Thus it is possible to have multiple ID type attributes on a single
     *   element, or ID type attributes on different elements which have the same value,
     *   or NMTOKEN type attributes which do not contain legal XML name tokens,
     *   and so forth.
     * </p>
     * 
     * @author Elliotte Rusty Harold
     *
     */
    public static final class Type {

        /**
         * <p>
         *   The type of attributes declared to have type CDATA
         *   in the DTD. The most general attribute type.
         *   All well-formed attribute values are valid for attributes of type CDATA.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type CDATA = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type ID
         *   in the DTD. In order to be valid, an ID type attribute
         *   must contain an XML name which is unique among other 
         *   ID type attributes in the document.
         *   Furthermore, each element may contain no more than one
         *   ID type attribute. However, XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type ID = new Type();
        
        /**
         * <p>
         *   The type of attributes declared to have type IDREF
         *   in the DTD. In order to be valid, an IDREF type attribute
         *   must contain an XML name which is also the value of  
         *   ID type attribute of some element in the document. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type IDREF = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type IDREFS
         *   in the DTD. In order to be valid, an IDREFS type attribute
         *   must contain a whote space separated list of
         *   XML names, each of which is also the value of  
         *   ID type attribute of some element in the document. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type IDREFS = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type NMTOKEN
         *   in the DTD. In order to be valid, a NMTOKEN type attribute
         *   must contain a single XML name token. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type NMTOKEN = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type NMTOKENS
         *   in the DTD. In order to be valid, a NMTOKENS type attribute
         *   must contain a white space separated list of  XML name tokens. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type NMTOKENS = new Type();


        /**
         * <p>
         *   The type of attributes declared to have type NOTATION
         *   in the DTD. In order to be valid, a NOTATION type attribute
         *   must contain the name of a notation declared in the DTD. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type NOTATION = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type ENTITY
         *   in the DTD. In order to be valid, a  ENTITY type attribute
         *   must contain the name of an unparsed entity declared in the DTD. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type ENTITY = new Type();

        /**
         * <p>
         *   The type of attributes declared to have type ENTITIES
         *   in the DTD. In order to be valid, an ENTITIES type attribute
         *   must contain a white space separated list of 
         *   names of unparsed entities declared in the DTD. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type ENTITIES = new Type();

        /**
         * <p>
         *   The type of attributes declared by an enumeration
         *   in the DTD. In order to be valid, a enumeration type attribute
         *   must contain exactly one of the names given in the enumeration 
         *   in the DTD. However, 
         *   XOM does not enforce
         *   such validity constraints.
         * </p>
         * 
         * <p>
         *   Most parsers report attributes of type enumeration as having type
         *   NMTOKEN. In this case, XOM will not distinguish NMTOKEN and
         *   enumerated attributes.
         * </p>
         * 
         * 
         * @author Elliotte Rusty Harold
         *
         */
        public static final Type ENUMERATION = new Type();
        
        /**
         * <p>
         *   The type of attributes not declared in the DTD.
         *   This type only appears in invalid documents.
         *   This is the default type for all attributes in
         *   documents without DTDs.
         * </p>
         * 
         * <p>
         *   Most parsers report attributes of undeclared 
         *   type as having type 
         *   CDATA. In this case, XOM will not distinguish CDATA and
         *   undeclared attributes.
         * </p>

         * @author Elliotte Rusty Harold
         *
         */
        public static final Type UNDECLARED = new Type();

        private Type() {}

    }

}
