// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;

import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.LocatorImpl;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * <p>
 * The <code>Builder</code> class is responsible for creating XOM Document objects 
 * from a URL, file, or input stream by reading an XML document.
 * A SAX parser is used to read the document and report any well-formedness errors.
 * </p>
 * 
 * <p>
 * need to test parsing multiple files in series????
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class Builder {

    private XMLReader parser;
    
    /**
     * <p>
     * This constructor uses the default factory and
     * chooses any available SAX2 parser in the following order:
     * </p>
     * 
     * <ol>
     * <li>Xerces 2.1.0</li>
     * <li>&AElig;lfred</li>
     * <li>Piccolo</li>
     * <li>Oracle</li>
     * <li>Crimson</li>
     * <li>The platform default specified by the org.xml.sax.driver system property</li>
     * </ol>
     * 
     * <p>
     * Parsers used must support the 
     * http://xml.org/sax/features/external-general-entities
     * and
     * http://xml.org/sax/features/external-parameter-entities
     * features XOM requires. Parsers that don't are rejected automatically. 
     * </p>
     * 
     * @throws XMLException if no satisfactory parser is installed in the local class path
     */
    public Builder() {       
        this(findParser(false), false);     
    }
    
    /**
     * <p>
     * This constructor creates a <code>Builder</code> based on a fully validating
     * parser. If the <code>validate</code> argument is true, then a validity error while
     * parsing will cause a fatal error; that is,
     *  it will throw a <code>ParseException</code>`.
     * </p>
     * 
     * @param validate true if the parser should validate the document while parsing
     *
     * @throws XMLException if no satisfactory parser is installed in the local class path
     */
    public Builder(boolean validate) {     
         this(findParser(validate), validate);     
    }

    private static String[] parsers = {
        "org.apache.xerces.parsers.SAXParser",
        "com.bluecast.xml.Piccolo",
        "gnu.xml.aelfred2.XmlReader",
        "oracle.xml.parser.v2.SAXParser",
        "org.apache.crimson.parser.XMLReaderImpl"
    };

    private static XMLReader findParser(boolean validate) {
        
        XMLReader parser = null;
        for (int i = 0; i < parsers.length; i++) {
            try { 
                parser = XMLReaderFactory.createXMLReader(parsers[i]);
                if (!validate) {
                    parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                    parser.setFeature("http://xml.org/sax/features/external-general-entities", true);
                    parser.setFeature("http://xml.org/sax/features/external-parameter-entities", true);
                }
                else {  
                    parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                    parser.setFeature("http://xml.org/sax/features/validation", true);
                    parser.setErrorHandler(new ValidityRequired());
                }               
                return parser;
            }
            catch (SAXException ex) {
                // Try the next one 
            }       
        }
        
        if (parser == null) {
            try { // default
                parser = XMLReaderFactory.createXMLReader();
                if (!validate) {
                    parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                    parser.setFeature("http://xml.org/sax/features/external-general-entities", true);
                    parser.setFeature("http://xml.org/sax/features/external-parameter-entities", true);
                }
                else {  
                    parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                    parser.setFeature("http://xml.org/sax/features/validation", true);
                    parser.setErrorHandler(new ValidityRequired());
                }               
            }
            catch (SAXException ex) {
                throw new XMLException("Could not find a SAX2 parser", ex);
            }
        }
        
        return parser;
        
    }    
    
    

    /**
     * <p>
     * Creates a new <code>Builder</code> based 
     * on the specified SAX parser <code>XMLReader</code>.
     * Custom SAX features and properties such as schema validation
     * can be set on this
     *  <code>XMLReader</code> before passing it to this method.
     * </p>
     * 
     * @param parser The SAX2 <code>XMLReader</code> that parses the document.
     * 
     * @throws XMLException if parser does not support the features XOM requires
     */ 
    public Builder(XMLReader parser) {
        this(parser, false);
    }
    

    /**
     * <p>
     * Creates a new Builder based 
     * on the specified parser object.
     * Custom SAX features and properties such as schema validation
     * can be set on this
     * XMLReader before passing it to this method.
     * </p>
     * 
     * <p>
     * If the validate param is true, then a validity error while
     * parsing will cause a fatal error; i.e. it will throw a ParseException
     * </p>
     * 
     * @param parser The SAX2 XMLReader that parses the document.
     * 
     * @param validate true if the parser should validate the document while parsing
     * 
     */ 
    public Builder(XMLReader parser, boolean validate) {
        this(parser, validate, new NodeFactory());
    }
    
    /**
     * <p>
     * This constructor creates a Builder based 
     * on the specified parser object.
     * Custom SAX features and properties such as schema validation
     * can be set on this
     * XMLReader before passing it to this method.
     * </p>
     * 
     * <p>
     * If the validate param is true, then a validity error while
     * parsing will cause a fatal error; i.e. it will throw a ParseException
     * </p>
     * 
     * @param parser The SAX2 XMLReader that parses the document.
     * @param validate true if the parser should validate the document while parsing
     * @param factory the NodeFactory this builder uses to create objects in the tree
     * 
     * @throws XMLException if parser does not support the features XOM requires
     * 
     */ 
    public Builder(XMLReader parser, boolean validate, NodeFactory factory) {
        if (!validate) {
            try { 
                parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                parser.setFeature("http://xml.org/sax/features/external-general-entities", true);
                parser.setFeature("http://xml.org/sax/features/external-parameter-entities", true);
            }
            catch (SAXNotRecognizedException ex) {
                throw new XMLException(parser.getClass().getName()
                 + " does not support the entity resolution features XOM requires.", ex);
            }
            catch (SAXNotSupportedException ex) {
                throw new XMLException(parser.getClass().getName() 
                 + " does not support the entity resolution features XOM requires.", ex);
            }
        }
        else {
            try { 
                parser.setFeature("http://xml.org/sax/features/namespace-prefixes", true);
                parser.setFeature("http://xml.org/sax/features/validation", true);
                parser.setErrorHandler(new ValidityRequired());
            }
            catch (SAXNotRecognizedException ex) {
                throw new XMLException(parser.getClass().getName() +
                 " does not support the validation features XOM requires.", ex);
            }
            catch (SAXNotSupportedException ex) {
                throw new XMLException(parser.getClass().getName() +
                 " does not support the validation features XOM requires.", ex);
            }
            
        }
        XOMHandler handler = new XOMHandler(factory);
        parser.setContentHandler(handler);
        try {
            parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
        }
        catch (SAXException ex) {
            // This parser does not support lexical events.
            // We can live without them, though it does mean
            // there won't be any comments or a doctype declaration in the tree.
        }
                
        try {
            parser.setProperty("http://xml.org/sax/properties/declaration-handler", handler);
        }
        catch (SAXException e) {
            // This parser does not support declaration events.
            // We can live without them, though it does mean
            // there won't be any internal DTD subset.
            // I do need to test the internal DTD subset accumulation????
        }
        
        this.parser = parser;
    }
    
    
    /**
     * <p>
     * This method parses the document at the specified URL.
     * </p>
     * 
     * @param systemID the URL (generally absolute) from which the document is read.
     *                   The URL's scheme must be one supported by the Java VM. 
     * 
     * @return the parsed <code>Document</code>
     * 
     * @throws ValidityException if a validity error is detected. This 
     *                             is only thrown if the builder has been
     *                             instructed to validate.
     * @throws ParseException  if a well-formedness error is detected
     * @throws IOException  if an I/O error such as a broken socket prevents the
     *                       document from being fully read.
     */
    public Document build(String systemID) 
      throws ParseException, IOException {

        try {
            parser.parse(systemID);
        }
        catch (SAXInvalidException ex) {
            // This should only be thrown if validation 
            // was requested from the constructor 
            throw new ValidityException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXParseException ex) {
           /* System.err.println(parser.getClass());
            System.err.println(parser.getErrorHandler());*/
            throw new ParseException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXException ex) {
           /* System.err.println(parser.getClass());
            System.err.println(parser.getErrorHandler());*/
            throw new ParseException(ex.getMessage(), ex);            
        } 
        XOMHandler handler = (XOMHandler) (parser.getContentHandler());
        return handler.getDocument();
        
    }

    /**
     * <p>
     * This method reads the document from an input stream.
     * </p>
     * 
     * @param in the <code>InputStream</code> from which the document is read.
     * 
     * @return Document the parsed <code>Document</code>
     * 
     * @throws ValidityException if a validity error is detected. This 
     *                             is only thrown if the builder has been
     *                             instructed to validate.
     * @throws ParseException  if a well-formedness error is detected
     * @throws IOException  if an I/O error such as a broken socket prevents the
     *                       document from being fully read.
     */
    public Document build(InputStream in) 
      throws ParseException, IOException {

        try {
            InputSource source = new InputSource(in);
            parser.parse(source);
        }
        catch (SAXInvalidException ex) {
            // This should only be thrown if validation 
            // was requested from the constructor 
            throw new ValidityException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXParseException ex) {
           /* System.err.println(parser.getClass());
            System.err.println(parser.getErrorHandler());*/
            throw new ParseException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXException ex) {
            throw new ParseException(ex.getMessage(), ex);          
        } 

        XOMHandler handler = (XOMHandler) (parser.getContentHandler());
        return handler.getDocument();
        
    }
    
    /**
     * <p>
     * This method reads the document from an input stream.
     * </p>
     * 
     * @param in the <code>Reader</code> from which the document is read. 
     * 
     * @return Document  the parsed <code>Document</code>
     * 
     * @throws ValidityException if a validity error is detected. This 
     *                             is only thrown if the builder has been
     *                             instructed to validate.
     * @throws ParseException  if a well-formedness error is detected
     * @throws IOException  if an I/O error such as a broken socket prevents the
     *                       document from being fully read.
     */
    public Document build(Reader in) 
      throws ParseException, IOException {

        try {
            InputSource source = new InputSource(in);
            parser.parse(source);
        }
        catch (SAXInvalidException ex) {
            // This should only be thrown if validation 
            // was requested from the constructor 
            throw new ValidityException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXParseException ex) {
           /* System.err.println(parser.getClass());
            System.err.println(parser.getErrorHandler());*/
            throw new ParseException(ex.getMessage(), ex.getLineNumber(), ex.getColumnNumber(), ex);            
        } 
        catch (SAXException ex) {
            throw new ParseException(ex.getMessage(), ex);          
        } 

        XOMHandler handler = (XOMHandler) (parser.getContentHandler());
        return handler.getDocument();
        
    }
    
    private static class ValidityRequired implements ErrorHandler {

        public void warning(SAXParseException exception) {
            // ignore warnings
        }
      
        public void error(SAXParseException exception) throws SAXInvalidException {           
            LocatorImpl locator = new LocatorImpl();
            locator.setLineNumber(exception.getLineNumber());
            locator.setColumnNumber(exception.getColumnNumber());
            throw new SAXInvalidException(exception.getMessage(),
             locator, exception);            
        }
      
        public void fatalError(SAXParseException exception) throws SAXParseException {
            throw exception;            
        }        
        
    }

}