// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 * <p>
 *   This class represents an XML comment such as <code>&lt;-- This is 
 *   a comment--></code>. It cannot have any child nodes.
 *   It can be a child of an <code>Element</code> or a <code>Document</code>.
 * It has essentially no internal substructure.
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class Comment extends LeafNode {

    private String data;

    /**
     * Creates a new Comment object from string data. The data is checked for 
     * legality according to XML 1.0 rules. Characters like the 
     * form feed and byte order mark are not allowed. Furthermore,
     * the two hyphen string "--" is not allowed and the last character of the 
     * comment may not be a hyphen.
     * 
     * @param data The initial text of the object
     */
    public Comment(String data) {
        setData(data);  
    }

    /**
     * Creates a new comment that's a copy of its argument.
     * The copy has the same data but no parent node.
     * 
     * @param comment The comment to copy
     */
    public Comment(Comment comment) {
        this.data = comment.data;  
    }

    /**
     * Sets the content of the Comment object to the specified string.
     * This string is checked for 
     * legality according to XML 1.0 rules. Characters that can be serialized
     * such as &lt; and &amp; are allowed. However, characters like the 
     * form feed and byte order mark are not allowed.
     * Furthermore, the string may not contain a double hyphen (<code>--</code>)
     * and may not end with a hyphen.
     * 
     * @param data The text to install in the object
     */
    public final void setData(String data) {
        Verifier.checkCommentData(data);
        check(data);
        if (data == null) data = "";
        // Is <!----> a legal comment. Yes it is. 
        this.data = data;
    }

    /**
     * Subclasses can override this method to perform additional checks.
     * However, this can only be used
     * to add checks, not remove them. All text in comments must be
     * potentially well-formed when serialized. 
     * 
     * @param data The text to check
     * 
     * @throws XMLException if the data does not satisfy the local constraints
     */
    protected void check(String data) {}

    /**
     * Returns the value of this comment as defined by XPath 1.0. 
     * The XPath string-value of a comment node is the string content of the node,
     * not including the initial &lt;-- and closing --&gt;.
     *  
     * @return The content of the comment.
     * @see nu.xom.Node#getValue()
     */
    public final String getValue() {
        return data;
    }

    // clones could bypass checks like this????
    // because strings are immutable
    private Comment() {}

    /**
     * <p>
     *   Returns a copy of this <code>Comment</code> object 
     *   which contains the same text, but does not have any parent.
     *   Thus, it can be inserted into a different document.
     * </p>
     * 
     */
    public final TreeNode copy()  {
        Comment comment = new Comment();
        comment.data = data;
        return comment;
    }

    /**
     * <p>
     *   Returns a <code>String</code> containing the actual XML
     *  form of the comment;
     *   for example, <code>&lt;--This is a comment--&gt;</code>. 
     * </p>
     * 
     * @return  a <code>String</code> containing a well-formed XML comment
     */
    public final String getStringForm() {
        return "<!--" + data + "-->";
    }


    /**
     * <p>
     *   Returns a string form of the comment suitable for debugging
     *   and diagnosis. It deliberately does not return an actual XML comment. 
     * </p>
     * 
     * 
     * @see java.lang.Object#toString()
     * @return  a representation of the object as a <code>String</code>
     */
    public final String toString() {
        return "[" + getClass().getName() + ": " + data + "]";
    }
    
}
