// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
// Many DOM interfaces such as Element and Document
// have name conflicts with XOM classes.
// Thus they cannot be imported, and this class
// must use their fully package qualified names.


/**
 * <p>
 * Converts XOM <code>Document</code> objects to and from DOM 
 * <code>Document</code> objects.
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 *
 */
public class DOMConverter {

    /**
     * <p>
     * Translates a DOM <code>org.w3c.dom.Document</code> object into
     * an equivalent <code>nu.xom.Document</code>.
     * The original DOM document is not changed.
     * Some DOM <code>Document</code> objects cannot 
     * be serialized as namespace well-formed XML, and thus cannot be 
     * converted to XOM.
     * </p>
     * 
     * @param domDocument the DOM document to translate
     * @return a XOM document
     * 
     * @throws XMLException if the DOM document is not a well-formed XML document
     */
    public static Document translate(org.w3c.dom.Document domDocument) {
        
        org.w3c.dom.Element domRoot = domDocument.getDocumentElement();
        Element xomRoot = convert(domRoot);
        Document xomDocument = new Document(xomRoot);
        
        org.w3c.dom.Node current = domDocument.getFirstChild();
        
        // prolog
        for (int position = 0; 
             current.getNodeType() != org.w3c.dom.Node.ELEMENT_NODE; 
             position++, current = current.getNextSibling()) {
            xomDocument.insertChild(position, convert(current));
        }
        // root element       
        current = current.getNextSibling();
        
        // epilog
        while (current != null) {
            xomDocument.appendChild(convert(current));
            current = current.getNextSibling();   
        }       
                       
        return xomDocument;
    }

    private static TreeNode convert(org.w3c.dom.Node node) {
        
        int type = node.getNodeType();
        switch (type) {
            case org.w3c.dom.Node.ELEMENT_NODE:
                return convert((org.w3c.dom.Element) node);   
            case org.w3c.dom.Node.COMMENT_NODE:
                return convert((org.w3c.dom.Comment) node);   
            case org.w3c.dom.Node.DOCUMENT_TYPE_NODE:
                return convert((org.w3c.dom.DocumentType) node);   
            case org.w3c.dom.Node.TEXT_NODE:
                return convert((org.w3c.dom.Text) node);
            case org.w3c.dom.Node.CDATA_SECTION_NODE:
                return convert((org.w3c.dom.Text) node);
            case org.w3c.dom.Node.PROCESSING_INSTRUCTION_NODE:
                return convert((org.w3c.dom.ProcessingInstruction) node);
            default:   
                return null;
        }
        
    }

    private static Comment convert(org.w3c.dom.Comment comment) {       
        return new Comment(comment.getNodeValue());
    }

    private static Text convert(org.w3c.dom.Text text) {       
        return new Text(text.getNodeValue());       
    }

    private static ProcessingInstruction convert(org.w3c.dom.ProcessingInstruction pi) {       
        return new ProcessingInstruction(pi.getTarget(), pi.getNodeValue());       
    }

    private static DocType convert(org.w3c.dom.DocumentType doctype) {
        
        DocType result = new DocType(doctype.getName(), doctype.getPublicId(), doctype.getSystemId());
        String internalSubset = doctype.getInternalSubset();
        if (internalSubset != null) result.setInternalDTDSubset(internalSubset);
        
        return result;
               
    }

    private static Element convert(org.w3c.dom.Element element) {       
        Element result = new Element(element.getTagName(), element.getNamespaceURI());

        
        // fill elements attributes and additional namespace declarations
        NamedNodeMap attributes = element.getAttributes();
        for (int i = 0; i < attributes.getLength(); i++) {
            org.w3c.dom.Attr attribute = (org.w3c.dom.Attr) attributes.item(i);
            String uri = attribute.getNamespaceURI();
            if (uri == null) uri = "";
            if (uri.equals("http://www.w3.org/2000/xmlns/")) {
                String name = attribute.getName();
                if (name.equals("xmlns")) continue;
                String prefix = name.substring(name.indexOf(':') + 1);
                String currentPrefix = result.getNamespaceURI(prefix);
                if (!prefix.equals(currentPrefix)) {
                    result.addNamespaceDeclaration(prefix, attribute.getValue());   
                }
            }    
            else {
                result.add(new Attribute(attribute.getName(), uri, attribute.getValue()));
                // add type????   
            }
        }
        
        // recurse children
        for (org.w3c.dom.Node current = element.getFirstChild(); current!= null; current = current.getNextSibling()) {
            result.appendChild(convert(current));
        }
        
        return result;       
    }

 
    /**
     * <p>
     * Translates a XOM <code>nu.xom.Document</code> object
     * into an equivalent <code>org.w3c.dom.Document</code> object.
     * The original XOM document is not changed.
     * Since DOM2 internal subsets are read-only,
     * the internal DTD subset is not converted. 
     * All other aspects of the document should be translated without
     * a problem.
     * </p>
     * 
     * @param document the XOM document to translate
     * 
     * @return a DOM document
     */
    public static org.w3c.dom.Document translate(Document document, 
      DOMImplementation impl) {

        Element root = document.getRootElement();
        String rootName = root.getQualifiedName();
        String rootNamespace = root.getNamespaceURI();
        DocType doctype = document.getDocType();     
        DocumentType domDOCTYPE = null;
        if (doctype != null) {
            domDOCTYPE = impl.createDocumentType(rootName, 
            doctype.getPublicID(), doctype.getSystemID());   
        }      
        
        org.w3c.dom.Document domDoc 
         = impl.createDocument(rootNamespace, rootName, domDOCTYPE);
        org.w3c.dom.Element domRoot = domDoc.getDocumentElement();
        
        boolean beforeRoot = true;
        for (int i = 0; i < document.getChildCount(); i++) {
            TreeNode original = document.getChild(i);
            // Need to test positioning of doctype
            if (original instanceof DocType) continue;
            else if (original instanceof Element) {
                convert((Element) original, domDoc);
                beforeRoot = false;   
            }
            else {
                org.w3c.dom.Node domNode = convert(original, domDoc);
                if (beforeRoot) domDoc.insertBefore(domNode, domRoot);
                else domDoc.appendChild(domNode);
            }
        } 
        
        return domDoc;
    }

    private static org.w3c.dom.Node convert(TreeNode node, org.w3c.dom.Document document) {
        
        org.w3c.dom.Node domNode;
        
        if (node instanceof Text) return convert((Text) node, document);
        else if (node instanceof Element) return convert((Element) node, document);
        else if (node instanceof Comment) return convert((Comment) node, document);
        else if (node instanceof ProcessingInstruction) return convert((ProcessingInstruction) node, document);
        else if (node instanceof Text) return convert((Text) node, document);
        else {
            throw new XMLException("Unexpected node type: " + node.getClass().getName());   
        }
         
    }

    private static org.w3c.dom.Comment convert(Comment comment, org.w3c.dom.Document document) {
        return document.createComment(comment.getValue());   
    }

    private static org.w3c.dom.Text convert(Text text, org.w3c.dom.Document document) {
        return document.createTextNode(text.getValue());        
    }

    private static org.w3c.dom.ProcessingInstruction convert(ProcessingInstruction pi, org.w3c.dom.Document document) {
        return document.createProcessingInstruction(pi.getTarget(), pi.getValue());        
    }

    private static org.w3c.dom.Element convert(Element element, org.w3c.dom.Document document) {
        
        org.w3c.dom.Element result;
        String namespace = element.getNamespaceURI();   
        String prefix = element.getNamespacePrefix();   
        Namespaces additional = element.getAdditionalNamespaces();
         
        if (element.getParent() instanceof Document) {
            result = document.getDocumentElement();   
        } 
        else if (namespace.equals("")) {
            result = document.createElement(element.getQualifiedName()); 
        }
        else {
            result = document.createElementNS(namespace, element.getQualifiedName());
        }
        
        
        addNamespaceAttributeIfNecessary(
            element,
            document,
            result,
            namespace,
            prefix);
        
        
        Attributes attributes = element.getAttributes();
        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = attributes.get(i);
            String attns = attribute.getNamespaceURI();
            Attr attr;
            if (attns.equals("")) {
                attr = document.createAttribute(attribute.getLocalName());
                result.setAttributeNode(attr);
            }
            else {
                attr = document.createAttributeNS(attns, attribute.getQualifiedName());
                result.setAttributeNodeNS(attr);
                String attPrefix = attribute.getNamespacePrefix();
                if (!namespace.equals(attns) ) {    
                    addNamespaceAttributeIfNecessary(element, document, result, attns, attPrefix);
                }
            }
            attr.setValue(attribute.getValue());
            // add type????

        }
        
        for (int i = 0; i < additional.size(); i++) {
            String additionalPrefix = additional.getPrefix(i);
            String uri = additional.getURI(additionalPrefix);
            Attr attr = document.createAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:" + additionalPrefix);
            result.setAttributeNodeNS(attr);
            attr.setValue(uri);
            // add type????
        }
        
        
        // children
        for (int i = 0; i < element.getChildCount(); i++) {
            result.appendChild(convert(element.getChild(i), document));   
        }
        
        return result;       
    }

    private static void addNamespaceAttributeIfNecessary(
        Element element,
        org.w3c.dom.Document document,
        org.w3c.dom.Element result,
        String namespace,
        String prefix) {
        boolean addNamespace = false;
        if (!"".equals(namespace)) {
            TreeNode parent = element.getParent();
            if (parent instanceof Document) addNamespace = true;
            else {
                Element parentElement = (Element) parent;
                if ( !(element.getNamespaceURI().equals(parentElement.getNamespaceURI(prefix))) ) {
                    addNamespace = true;   
                }   
            }   
            
        }
        if (addNamespace) {
            Attr xmlns;
            if ("".equals(prefix)) {
                xmlns = document.createAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns");
            }
            else {
                xmlns = document.createAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:" + prefix);                
            } 
            xmlns.setValue(namespace);
            result.setAttributeNodeNS(xmlns);   
        }
    }


}
