// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;


/**
 * <p>
 *   This class represents an XML document type declaration such as 
 * </p>
 * 
 * <p><code>&lt;!DOCTYPE book SYSTEM "dobookx.dtd"></code></p> 
 * 
 * <p>
 *   Note that this is not the same thing as a document
 * type <em>definition</em> (DTD). XOM does not currently model the DTD.
 * The document type declaration contains or points to the
 * DTD, but it is not the DTD.
 * </p>
 * 
 * <p>
 * A <code>DocType</code> object does not have any child nodes.
 * It can be a child of a <code>Document</code>.
 * </p>
 * 
 * <p>
 * Each <code>DocType</code> object has four properties, some of which may be null:
 * </p>
 * 
 * <ul>
 *   <li>The declared name of the root element (which 
 * does not necessarily match the actual root element name
 *  in an invalid document.)</li>
 *   <li>The public identifier (which may be null)</li>
 *   <li>The system identifier (which may be null)</li>
 *   <li>The internal DTD subset as a <code>String</code>
 *       (which may be null).</li>
 * 
 * </ul>
 * 
 * <p>
 *   The first three properties are read-write.
 *   The internal DTD subset is read-only. 
 *   XOM fills it in when a document is read by a parser.
 *   However, it cannot be changed, because XOM cannot 
 *   currently check that an internal DTD subset is well-formed.
 *   This restriction may be relaxed in a future version. 
 * </p>
 * 
 *
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class DocType extends LeafNode {

    private String rootName;
    private String systemID;
    private String publicID;
    // Internal DTD subset purely for parsing
    private StringBuffer internalDTDSubset = new StringBuffer();

    /**
     * Creates a new document type declaration with a system ID
     * and a public ID.
     * It has the general form <code>&lt;!DOCTYPE rootElementName PUBLIC 
     * "publicID" "systemID"&gt;</code>.
     * 
     * @param rootElementName the name specified for the root element
     * @param publicID the public ID of the external DTD subset
     * @param systemID the URL of the external DTD subset
     * 
     * @throws IllegalNameException  if <code>rootElementName</code> is not a legal XML 1.0 name
     * @throws IllegalDataException if <code>systemID</code> is not a legal XML 1.0 system literal
     *                                or <code>publicID</code> is not a legal XML 1.0 public identifier
     */
    public DocType(String rootElementName, String publicID, String systemID) {
        setRootElementName(rootElementName);    
        if (systemID != null && !"".equals(systemID)) setSystemID(systemID);  
        if (publicID != null && !"".equals(publicID)) setPublicID(publicID);    
    }

    /**
     * Creates a new document type declaration with a system ID
     * but no public ID.
     * It has the general form <code>&lt;!DOCTYPE rootElementName SYSTEM "systemID"&gt;</code>.
     * 
     * @param rootElementName the name specified for the root element
     * @param systemID the URL of the external DTD subset
     * 
     * @throws IllegalNameException  if the rootElementName is not a legal XML 1.0 name
     * @throws IllegalDataException if the system ID is not a legal XML 1.0 system literal
     */
    public DocType(String rootElementName, String systemID) {
        setRootElementName(rootElementName);    
        setSystemID(systemID);  
    }

    /**
     * Creates a new document type declaration with no public or system ID.
     * It has the general form <code>&lt;!DOCTYPE rootElementName &gt;</code>.
     * 
     * @param rootElementName the name specified for the root element
     * @throws IllegalNameException  if the rootElementName is not a legal XML 1.0 name
     */
    public DocType(String rootElementName) {
        setRootElementName(rootElementName);    
    }

    /**
     * Creates a new <code>DocType</code> that's a copy of its argument.
     * The copy has the same data but no parent document.
     * 
     * @param doctype The DocType to copy
     */
    public DocType(DocType doctype) {
        this.internalDTDSubset = new StringBuffer(doctype.internalDTDSubset.toString());
        this.publicID = doctype.publicID;
        this.systemID = doctype.systemID;
        this.rootName = doctype.rootName;    
    }


    /**
     * Sets the name the document type declaration specifies for the 
     * root element. In an invlaid document, this may 
     * not be the same as the actual root
     * element name.
     * 
     * @param name the root element name given by the document type declaration.
     */
    public final void setRootElementName(String name) {
        Verifier.checkXMLName(name);
        checkRootElementName(name);
        this.rootName = name;
    }

    /**
     * Returns the name the document type declaration specifies for the 
     * root element. In an invlaid document, this may 
     * not be the same as the actual root
     * element name.
     * 
     * @return  the declared name of the root element
     */
    public final String getRootElementName() {
        return rootName;
    }

    /**
     * Returns complete internal DTD subset in a single string.
     * White space may not be preserved completely accurately,
     * but all declarations should be in place. 
     * 
     * @return  the internal DTD subset
     */
    public final String getInternalDTDSubset() {
        return internalDTDSubset.toString();
    }
    
    final void addDeclaration(String declaration) {
        internalDTDSubset.append(declaration);  
        internalDTDSubset.append("\r\n");   
    }

    final void addComment(String comment) {
        internalDTDSubset.append(comment);  
        internalDTDSubset.append("\r\n");   
    }

    final void addProcessingInstruction(String instruction) {
        internalDTDSubset.append(instruction);  
        internalDTDSubset.append("\r\n");   
    }

    final void setInternalDTDSubset(String internalSubset) {
        this.internalDTDSubset = new StringBuffer(internalSubset);   
    }


    /**
     * Sets the public ID for the external DTD subset.
     * This can only be set after a system ID has been set, because
     * XML requires that all document type declarations with public IDs have system IDs.
     * Passing null removes the public ID. 
     * 
     * @param id the public identifier of the external DTD subset
     * 
     * @throws IllegalDataException if the public ID does not satisfy the rules for public IDs in XML 1.0
     * @throws WellformednessException if no system ID has been set
     */
    public final void setPublicID(String id) {  
        if (systemID == null) {
            throw new WellformednessException("Cannot have a public ID without a system ID");   
        }
        Verifier.checkPublicID(id);
        checkPublicID(id);
        this.publicID = id;
    }

    /**
     * Returns the public ID of the external DTD subset. 
     * This is null if there is no external DTD subset
     * or if it does not have a public identifier.
     * 
     * @return  the public ID of the external DTD subset.
     */
    public final String getPublicID() { 
        return publicID;
    }

    /**
     * Sets the system ID for the external DTD subset.
     * This is normally a relative or absolute URL.
     * Passing null removes the system ID, but only if the 
     * public ID has been removed first. Otherwise,
     * passing null causes a WellformednessException.
     * 
     * @param id the URL of the external DTD subset
     * 
     * @throws IllegalDataException if the system ID is not a legal XML 1.0 system literal
     * @throws WellformednessException if the public ID is non-null and you attempt to
     *                                   remove the system ID
     */
    public final void setSystemID(String id) {
        if (id == null && publicID != null) {
            throw new WellformednessException("Cannot remove system ID without removing public ID first");   
        }
        Verifier.checkSystemLiteral(id);
        checkSystemID(id);
        this.systemID = id;
    }
    
    /**
     * Returns the system ID of the external DTD subset. 
     * This is a URL. It is null if there is no external DTD subset.
     * 
     * @return  the URL for the external DTD subset.
     */
    public final String getSystemID() { 
        return systemID;
    }


    /**
     * Subclasses can override this method to perform additional checks
     * on the root element name beyond what XML 1.0 requires.
     * For example, an HTMLDocType subclass might throw an exception
     * if any name other than "html" were passed to this method.
     * However, this can only be used
     * to add checks, not remove them. All document type declarations must be
     * potentially well-formed when serialized. 
     * 
     * @param name The root element name specified by the document type declaration.
     * 
     * @throws XMLException if the proposed root element name does not satisfy the local constraints
     */
    protected void checkRootElementName(String name) {}

    /**
     * Subclasses can override this method to perform additional checks
     * on the system ID beyond what XML 1.0 requires.
     * For example, it might wish to check that the URL scheme is recognized.
     * However, this can only be used
     * to add checks, not remove them. All document type declarations must be
     * potentially well-formed when serialized. 
     * 
     * @param systemID The URL of the external DTD subset.
     * 
     * @throws XMLException if the proposed system ID does not satisfy the local constraints
     */
    protected void checkSystemID(String systemID) {}

    /**
     * Subclasses can override this method to perform additional checks
     * on the system ID beyond what XML 1.0 requires.
     * For example, an <code>HTMLDocType</code> subclass might check that the public ID
     * were one of the three IDs defined by the XHTML 1.0 specification.
     * However, this can only be used
     * to add checks, not remove them. All document type declarations must be
     * potentially well-formed when serialized. 
     * 
     * @param publicID The proposed public ID for the external DTD subset.
     * 
     * @throws XMLException if the proposed public ID does not satisfy the local constraints
     */
    protected void checkPublicID(String publicID) {}

    /**
     * Returns the empty string.
     * XPath 1.0 does not define a value for document type declarations.
     * 
     * 
     * @see nu.xom.Node#getValue()
     */
    public final String getValue() {
        return "";
    }


    /**
     * <p>
     *   This method returns a string form of the 
     * <code>DocType</code> suitable for debugging
     *   and diagnosis. It deliberately does not return an actual XML document
     * type declaration. 
     * </p>
     * 
     * 
     * @see java.lang.Object#toString()
     * @return  a string representation of the object
     */
    public final String toString() {
        return "[DocType: " + rootName + "]";
    }

    /**
     * <p>
     *   This method returns a copy of this <code>DocType</code> 
     *   which has the same system ID, public ID, root element name,
     *   and internal DTD subset, but does not belong to a document.
     *   Thus, it can be inserted into a different document.
     * </p>
     * 
     * @see nu.xom.TreeNode#copy()
     */
    public final TreeNode copy() {      
        return new DocType(this);
    }

    /**
     * <p>
     *  Returns a string containing the actual XML
     *  form of the document type declaration represented by this 
     *   object;
     *   for example, <code>&lt;!DOCTYPE book SYSTEM "docbookx.dtd"></code>. 
     * </p>
     * 
     * @return  a <code>String</code> containing an XML document type declaration
     */
    public final String getStringForm() {   
        StringBuffer result = new StringBuffer();
        result.append("<!DOCTYPE ");
        result.append(rootName);
        if (publicID != null) {
          result.append(" PUBLIC \"" + publicID + "\" \"" + systemID + "\"");
        } 
        else if (systemID != null) {
          result.append(" SYSTEM \"" + systemID + "\"");
        } 
        
        String internalDTDSubset = this.internalDTDSubset.toString();
        if (!internalDTDSubset.equals("")) {
            result.append(" [\r\n");    
            result.append(internalDTDSubset);   
            result.append("]"); 
        }
        
        result.append(">");
        return result.toString();
    }

}