// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 *  <p>
 * The <code>Document</code> class represents
 * a complete XML document including its root element,
 * prolog, and epilog.
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class Document extends TreeNode {

    /**
     * <p>
     * Creates a new <code>Document</code> object with the
     * specified root element.
     * </p>
     * 
     * @param root the root element of this document
     */
    public Document(Element root) {
        checkRoot(root);
        super.insertChild(0, root);
    }

    /**
     * <p>
     * Creates a copy of the <code>Document</code>.
     * </p>
     * 
     * @param doc the Document to copy
     */
    public Document(Document doc) {

      super.insertChild(0, doc.getRootElement().copy());
      for (int i = 0; i < getChildCount(); i++) {
           TreeNode child = getChild(i);
           if (!(child instanceof Element)) {
              this.insertChild(i, child.copy());
           }
        }

    }


    /**
     * <p>
     * Inserts a child node at the specified position.
     * Inserting at position 0 makes the child the first child
     * of this node. Inserting at the position <code>howManyChildren()</code>
     * makes the child the last child of the node.
     * </p>
     * 
     * @param position where to insert the child
     * @param child the node to insert
     * 
     * @throws IllegalAddException if child is not a Comment, ProcessingInstruction, or DocType
     * @throws MultipleParentException if child already has a parent
     * @throws NullPointerException if child is null
     * @throws IndexOutOfBoundsException if the position is negative or greater than
     *                                     the number of children of the node.`
     */
    public final void insertChild(int position, TreeNode child) {
        if (child instanceof Comment || child instanceof ProcessingInstruction) {
            super.insertChild(position, child);
        }
        else if (child instanceof DocType) {
            if (position <= getRootPosition()) {
                // need to remove existing doctype
                DocType oldDocType = getDocType(); 
                super.insertChild(position, child);
                if (oldDocType != null) oldDocType.detach();
            }
            else {
                throw new IllegalAddException(
                    "Cannot add a document type declaration after the root element");               
            }
        }
        else if (child instanceof Element) {
            throw new IllegalAddException("Cannot add a second root "
             + " element to a Document.");
        }
        else {
            throw new IllegalAddException("Cannot add a "
             + child.getClass() + " to a Document.");
        }

    }

    private int getRootPosition() {
        for (int i = 0; i < getChildCount(); i++) {
             Node child = getChild(i);
             if (child instanceof Element) {
                return i;
             }
         }
        // It should not be possible to get here
        throw new WellformednessException("Missing root element");
    }

    /**
     * Returns this document's document type declaration, or null if
     * it doesn't have one.
     *
     * @return DocType the document type declaration
     */
    public final DocType getDocType() {
        for (int i = 0; i < getChildCount(); i++) {
             Node child = getChild(i);
             if (child instanceof DocType) {
                return (DocType) child;
             }
         }
         return null;
    }


    /**
     * Returns this document's root element.
     * This is guaranteed to be non-null.
     *
     * @return Element the root element
     */
    public final Element getRootElement() {
        for (int i = 0; i < getChildCount(); i++) {
             Node child = getChild(i);
             if (child instanceof Element) {
                return (Element) child;
             }
         }
         // It should not be possible to get here
         throw new WellformednessException("Missing root element");
    }

    /**
     * Replaces the current root element
     * with a different root element.
     * 
     * @param root the new root element
     * 
     * @throws MultipleParentException if root has a parent
     * @throws XMLException if root is not legal for this subclass of Document
     * @throws NullPointerException if root is null; verify with test????
     */
    public final void setRootElement(Element root) {
        checkRoot(root);
        super.insertChild(getRootPosition(), root);
        super.removeChild(getRootPosition()+1);
    }

    /**
     * Subclasses can override this method to perform additional checks
     * on the root element beyond what XML 1.0 requires.
     * For example, an <code>XHTMLDocument</code> subclass might throw an exception
     * if the proposed root element were not an <code>html</code> element.
     * 
     * @param root The new root element.
     * 
     * @throws XMLException if the proposed root element does not satisfy the local constraints
     */
    protected void checkRoot(Element root) {}

    /**
     * Removes the child of this document at the specified position.
     * Indexes begin at 0 and count up to one less than the number
     * of children of this document. The root element cannot be removed.
     * Instead, use <code>setRootElement</code> to replace the
     * existing root element with a different element.
     * 
     * @param position index of the node to remove
     * 
     * @throws IndexOutOfBoundsException if the index is negative or greater than
     *                                     the number of children of this document - 1.
     * @throws WellformednessException if the index points to the root element.
     */
    public final void removeChild(int position) {
        if (position == getRootPosition()) {
            throw new WellformednessException("Cannot remove the root element");
        }
        super.removeChild(position);
    }

    /**
     * Removes the specified child from this node.
     * It throws a <code>NoSuchChildException</code> 
     * if the node is not a child of this node. The root element cannot be removed.
     * Instead, use <code>setRootElement</code> to replace the
     * existing root element with a different element.
     * 
     * @param child node to remove
     * 
     * @throws NoSuchChildException if the node is not a child of this node.
     * @throws WellformednessException if child is the root element.
     */
    public final void removeChild(TreeNode child) {
        if (child == getRootElement()) {
            throw new WellformednessException("Cannot remove the root element");
        }
        super.removeChild(child);
    }

    /**
     * Throws a <code>WellformednessException</code> because the root element cannot be removed.
     * 
     * @throws WellformednessException because the root element cannot be removed.
     */
    public final void removeChildren() {
        throw new WellformednessException("Cannot remove the root element");
    }


    /**
     * Returns the value of the document as defined by XPath 1.0.
     * This is the same as the value of the root element,
     * which is the complete PCDATA
     * content of the root element, without any tags, comments, or 
     * processing instructions after all entity and character
     * references have been resolved.
     * 
     * @see nu.xom.Node#getValue()
     * 
     * @return  value of the root element of this document
     * 
     */
    public final String getValue() {
        return getRootElement().getValue();
    }

    /**
     * Returns the actual complete, well-formed XML document as a <code>String</code>.
     * Significant white space is preserved. Insignificant white space in tags,
     * the prolog, the epilog, and the internal DTD subset is not preserved.
     * Entity and character references are not preserved. 
     * The entire document is contained in this one string.
     * 
     * @return  A <code>String</code> containing this entire  XML document
     */
    public final String getStringForm() {
    
        StringBuffer result = new StringBuffer();

        // XML declaration
        result.append("<?xml version=\"1.0\"?>\r\n");
        
        // children
        for (int i = 0; i < getChildCount(); i++) {
            result.append(getChild(i).getStringForm());
            result.append("\r\n");  
        }
        
        return result.toString();
        
    }

    
    /**
     * Returns a complete copy of this document.
     * 
     * @see nu.xom.TreeNode#copy()
     */
    public TreeNode copy() {
        return new Document(this);
    }


    /**
     * Returns a string representation of this node suitable for
     * debugging and diagnosis. This is <em>not</em>
     * the XML representation of this document.
     * 
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        return "[" + getClass() + ": " + getRootElement().getQualifiedName() + "]"; 
    }

}