// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import java.util.ArrayList;


/**
 * <p>
 * Represents an XML element. Each element has the following properties:
 * </p>
 * 
 * <ul>
 *   <li>Local name</li>
 *   <li>Prefix (which may be null or the empty string) </li>
 *   <li>Namespace URI (which may be null)</li>
 *   <li>A list of attributes</li>
 *   <li>A list of additional namespace declarations</li>
 *   <li>A list of child nodes</li>
 * </ul>
 * 
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 *
 */
public class Element extends TreeNode {

    private String localName;
    private String prefix;
    private String URI;

    private final Attributes attributes = new Attributes();
    private final Namespaces namespaces = new Namespaces();


    /**
     * Creates a new element in no namespace.

     * @param name the name of the element
     * 
     * @throws IllegalNameException  if name is not a legal XML 1.0 non-colonized name
     */
    public Element(String name) {
        this(name, "");
    }

    /**
     * Creates a new element in the specified namespace.

     * @param name the qualified name of the element
     * @param uri the namespace URI of the element
     * 
     * @throws IllegalNameException  if name is not a legal XML 1.0 non-colonized name
     * @throws IllegalDataException if uri is not an RFC 2396 URI reference
     */
    public Element(String name, String uri) {
        
        prefix = "";
        String localName = name;
        if (name.indexOf(':') >= 0) {
            prefix = name.substring(0, name.indexOf(':'));   
            localName = name.substring(name.indexOf(':') + 1);   
        }
        
        setNamespacePrefix(prefix);
        setLocalName(localName);
        setNamespaceURI(uri);
    }

    /**
     * Creates a copy of the specified element.

     * @param element the element to copy
     * 
     */
    public Element(Element element) {
        this(element.getQualifiedName(), element.URI);
        
        for (int i = 0; i < element.getChildCount(); i++) {
           TreeNode child = element.getChild(i);
           this.appendChild((TreeNode) child.copy());
        }
        
        // Attach additional namespaces
        for (int i = 0; i < element.namespaces.size(); i++) {
            String prefix = element.namespaces.getPrefix(i);
            String uri = element.namespaces.getURI(prefix);
            this.addNamespaceDeclaration(prefix, uri);
        }
        
        // Attach clones of attributes
        for (int i = 0; i < element.attributes.size(); i++) {
            Attribute a = element.attributes.get(i);
            this.add(new Attribute(a)); 
        }
        
    }


    /**
     * Returns a list of the immediate child elements of this element
     * with the specified name in no namespace.
     * 
     * @param name the name of the elements included in the list
     * 
     * @return Elements
     */
    public final Elements getChildElements(String name) {
        return getChildElements(name, "");
    }

    /**
     * Returns a list of the immediate child elements of this element
     * with the specified local name and namespace URI.
     * 
     * @param localName the name of the elements included in the list
     * @param namespace the namespace URI of the elements included in the list
     * 
     * @return Elements
     */
    public final Elements getChildElements(String localName, 
     String namespace) {
        
        Elements elements = new Elements();
        for (int i = 0; i < getChildCount(); i++) {
            TreeNode child = getChild(i);
            if (child instanceof Element) {
                Element element = (Element) child;
                if (localName.equals(element.getLocalName())
                  && namespace.equals(element.getNamespaceURI())) {
                    elements.add(element);
                }   
            }
        }
        return elements;    
        
    }

    /**
     * Returns a list of all the immediate child elements of this element.
     * 
     * @return Elements
     */
    public final Elements getChildElements() {
        
        Elements elements = new Elements();
        for (int i = 0; i < getChildCount(); i++) {
            TreeNode child = getChild(i);
            if (child instanceof Element) {
                Element element = (Element) child;
                elements.add(element);  
            }
        }
        return elements;    
        
    }

    // should these be getOnlyChild() and perhaps throw an exception if it isn't the
    // only child????

    /**
     * Returns the first child
     * element with the specified name in no namespace.
     * If there is no such element, it returns null.
     * 
     * @param name the name of the element to return
     * @return Element the first child with the specifed local name in no namespace.
     * or null if there is no such element.
     */
    public final Element getFirstChildElement(String name) {
        return getFirstChildElement(name, "");
    }

    /**
     * Returns the first child
     * element with the specified local name and namespace URI.
     * If there is no such element, it returns null.
     * 
     * @param localName the local name of the element to return
     * @param namespace the namespace URI of the element to return
     * 
     * @return Element the first child with the specifed local name in no namespace.
     * or null if there is no such element.
     */
    public final Element getFirstChildElement(String localName, 
     String namespace) {
        
        for (int i = 0; i < getChildCount(); i++) {
            TreeNode child = getChild(i);
            if (child instanceof Element) {
                Element element = (Element) child;
                if (localName.equals(element.getLocalName())
                  && namespace.equals(element.getNamespaceURI())) {
                    return element;
                }   
            }
        }
        return null;    
        
    }



    /**
     * Adds an attribute to this element, replacing any existing 
     * attribute with the same name.
     * 
     * @param attribute the attribute to add
     * 
     * @throws IllegalAddException if the attribute already has a parent
     * @throws NamespaceException - if the attribute is in a namespace which 
     *                              conflicts with the element's own namespace
     *                              or the namespace of another attribute
     *                              or an additional namespace declaration
     * @throws XMLException if a subclass has rejected this attribute
     */
    // Should this be called setAttribute????
    public final void add(Attribute attribute) {

        if (attribute.getParent() != null) {
            throw new IllegalAddException("Attribute already has a parent");
        }
        
        // check for namespace conflicts
        if (!attribute.getNamespacePrefix().equals("")) {
            if (prefix.equals(attribute.getNamespacePrefix())
              && !(getNamespaceURI().equals(attribute.getNamespaceURI()))) {
                throw new NamespaceException("Prefix of " + attribute.getQualifiedName() 
                + " conflicts with element prefix " + prefix);  
            }
            // check for conflicts with additional namespaces
            String existing = namespaces.getURI(attribute.getNamespacePrefix());
            if (existing != null && !existing.equals(attribute.getNamespaceURI())) {
                throw new NamespaceException("Attribute prefix  " + attribute.getNamespacePrefix() 
                + " conflicts with namespace declaration.");            
            }
            // check for conflicts with the namespaces of other attributes
            for (int i = 0; i < attributes.size(); i++) {
                Attribute a = attributes.get(i);
                if (a.getNamespacePrefix().equals(attribute.getNamespacePrefix())) {
                    if (a.getNamespaceURI().equals(attribute.getNamespaceURI())) break;
                    else {
                        throw new NamespaceException(
                         "Prefix of " + attribute.getQualifiedName() 
                         + " conflicts with " + a.getQualifiedName());
                    }
                }   
            }
        }
        
        checkAttribute(attribute);
        attributes.add(attribute);
        attribute.setParent(this);
        
    }

    /**
     * Subclasses can override this method to perform additional checks
     * on the specific attributes added to this element beyond what XML 1.0 requires.
     * For example, an HTMLSpan subclass might throw an exception
     * if any attribute were passed to this method which was not allowed
     * on an HTML span element.
     * 
     * @param attribute The attribute to check.
     * 
     * @throws XMLException if the proposed attribute does not satisfy the local constraints
     */
    protected void checkAttribute(Attribute attribute) {}

    // Does subclassing require a check????
    /**
     * Removes an attribute from this element.
     * 
     * @param attribute the attribute to remove
     * 
     * @throws XMLException if this element is not the parent of attribute
     * 
     */
    public final void removeAttribute(Attribute attribute) {
        attributes.remove(attribute);
        attribute.setParent(null);
    }

    /**
     * Returns the attribute with the specified name in no namespace,
     * or null if this element does not have an attribute 
     * with that name.
     * 
     * @param name the name of the attribute
     * 
     * @return  the attribute of this element with the specified name
     */
    public final Attribute getAttribute(String name) {
        return getAttribute(name, "");
    }

    /**
     * Returns the attribute with the specified name and namespace URI,
     * or null if this element does not have an attribute 
     * with that name in that namespace.
     * 
     * @param name the name of the attribute
     * @param namespaceURI the namespace of the attribute
     * 
     * @return  the attribute of this element with the specified name and namespace
     */
    public final Attribute getAttribute(String localName, String namespaceURI) {
        return attributes.get(localName, namespaceURI);
    }
    
    /**
     * Returns a read-only list of the attributes of this element.
     * The list is comatose; that is, changes to the element are not reflected
     * in the contents of the list. However, changes to the individual attributes
     * are reflected.
     * 
     * @return Attributes
     */
    public final Attributes getAttributes() {
        Attributes copy = new Attributes();
        copy.attributes = (ArrayList) (attributes.attributes.clone()); 
        return copy;    
    }

    /**
     * Returns the value of the attribute with the specified name in no namespace,
     * or null if this element does not have an attribute 
     * with that name.
     * 
     * @param name the name of the attribute
     * 
     * @return the value of the attribute of this element with the specified name
     */
    public final String readAttribute(String name) {
        return readAttribute(name, "");
    }

    /**
     * Returns the value of the attribute with the specified name and namespace URI,
     * or null if this element does not have such an attribute.
     * 
     * @param name the name of the attribute
     * @param namespaceURI the namespace of the attribute
     * 
     * @return the value of the attribute of this element with the specified name and namespace
     */
    public final String readAttribute(String localName, String namespaceURI) {
        Attribute attribute = attributes.get(localName, namespaceURI);
        if (attribute == null) return null;
        else return attribute.getValue();
    }
    
    /**
     * Returns the local name of this element, not including the 
     * namespace prefix or colon.
     * 
     * @return  the local name of this element
     */
    public final String getLocalName() {
        return localName;
    }

    /**
     * Returns the complete name of this element, including the 
     * namespace prefix if this element has one.
     * 
     * @return  the qualified name of this element
     */
    public final String getQualifiedName() {
        if (prefix.equals("")) return localName;
        else return prefix + ":" + localName;
    }

    /**
     * Returns the namespace URI of this element.
     * 
     * @return  the namespace URI of this element
     */
    public final String getNamespaceURI() {
        return URI;
    }

    /**
     * Returns the namespace URI mapped to the specified
     * prefix within this element. Returns null if this prefix
     * is not associated with a URI.
     * 
     * @return the namespace URI mapped to <code>prefix</code>
     */
    public final String getNamespaceURI(String prefix) {
        String result = getLocalNamespaceURI(prefix);
        if (result == null && getParent() instanceof Element) {
            Element parent = (Element) getParent();
            result = parent.getLocalNamespaceURI(prefix);
        }
        
        return result;
    }

    private String getLocalNamespaceURI(String prefix) {
        
        if (this.prefix.equals(prefix)) return this.URI;
        
        if (prefix.equals("xml")) return "http://www.w3.org/XML/1998/namespace";
        // This next line uses the original Namespaces 1.0 specification.
        // Namespaces 1.0 + errata is different
        if (prefix.equals("xmlns")) return "";
        // Look in the additional namespace declarations
        String result = namespaces.getURI(prefix);
        if (result != null) return result;
        
        // Look in the attributes
        for (int i = 0; i < attributes.size(); i++) {
            Attribute a = attributes.get(i);
            if (a.getNamespacePrefix().equals(prefix)) {
                return a.getNamespaceURI();
            }   
        }       
        
        return null;
        
    }

    /**
     * Returns the prefix of this element, or null
     * if this element does not have a prefix.
     * 
     * @return the prefix of this element
     */
    public final String getNamespacePrefix() {
        return prefix;
    }

    
    // does this need a check method????
    /**
     * Sets the local name of this element.
     * 
     * @param localName the new local name
     * 
     * @throws IllegalNameException  if <code>localName</code> is not
     * a legal, non-colonized name
     */
    public final void setLocalName(String localName) {
        Verifier.checkNCName(localName);
        this.localName = localName;
    }

    // does this need a check method????
    /**
     * Sets the namespace URI of this element.
     * 
     * @param URI the new namespace URI
     * 
     * @throws IllegalDataException if <code>URI</code> is not an RFC2396 URI reference
     */
    public final void setNamespaceURI(String URI) {
        if (URI == null) URI = "";
        Verifier.checkNamespaceURI(URI);
        this.URI = URI;
        
        // do I need to check how this affects or conflicts with
        // attribute namespaces and additional
        // namespace declarations????
        
    }

    // does this need a check method????
    /**
     * Sets the namespace prefix of this element.
     * 
     * @param prefix - the new namespace prefix
     * 
     * @throws IllegalNameException  if <code>prefix</code> is not a legal XML non-colonized name
     */
    public final void setNamespacePrefix(String prefix) {
        if (prefix == null  || prefix.equals("")) prefix = "";
        else Verifier.checkNCName(prefix);
        this.prefix = prefix;
        // do I need to check how this affects or conflicts with
        // attribute namespaces and additional
        // namespace declarations????
    }

    /**
     * <p>
     * Inserts a child node at the specified position.
     * Inserting at position 0 makes the child the first child
     * of this node. Inserting at the position <code>howManyChildren()</code>
     * makes the child the last child of the node.
     * </p>
     * 
     * <p>
     * All the other methods that add a node to the tree,
     * invoke this method ultimately.
     * </p>
     * 
     * @param position where to insert the child
     * @param child the node to insert
     * 
     * @throws IllegalAddException if <code>child</code> is a <code>Document</code>
     * @throws MultipleParentException if <code>child</code> already has a parent
     * @throws NullPointerException if <code>child</code> is null
     * @throws IndexOutOfBoundsException if the position is negative or greater than
     *                                     the number of children of this element.
     */
    public final void insertChild(int position, TreeNode child) {
        if (child instanceof LeafNode || child instanceof Element) {
            super.insertChild(position, child);
        }
        else {
            throw new IllegalAddException("Cannot add a "
             + child.getClass().getName() + " to an Element.");
        }

    }

    /**
     * Declares a namespace prefix. This is only necessary when
     * prefixes are used in element content and attribute values,
     * as in XSLT and the W3C XML Schema Language. Do not use this method 
     * to declare prefixes for element and attribute names.
     * 
     * @param prefix - the prefix to declare
     * @param URI the URI to map the prefix to
     * 
     * @throws IllegalDataException if <code>URI</code> is not an RFC2396 URI reference
     * @throws IllegalNameException  if <code>prefix</code> is not a legal XML non-colonized name
     * @throws NamespaceException - if the mapping conflicts with an existing element, attribute,
     *                              or additional namespace declaration
     */
    public final void addNamespaceDeclaration(String prefix, String URI) {

        if (prefix == null) prefix = "";
        if (URI == null) URI = "";
        
        Verifier.checkNCName(prefix);
        Verifier.checkNamespaceURI(URI);
        
        // check to see if this is the xmlns or xml prefix
        if (prefix.equals("xmlns")) {
            throw new NamespaceException("The xmlns prefix cannot bound to any URI");   
        }
        else if (prefix.equals("xml")) {
            if (URI.equals("http://www.w3.org/XML/1998/namespace")) {
                // This is done automatically
                return; 
            }
            else {
                throw new NamespaceException("Wrong namespace URI for xml prefix: " + URI);
            }   
        }
        
        
        // check for conflicts with element prefix
        if (prefix.equals(this.prefix)) {
            if (URI.equals(this.URI)) return;
            else {
                throw new NamespaceException(
                  "Additional namespace conflicts with element namespace."
                );  
            }
        }
        
        // check for conflicts with attribute prefixes
        for (int i = 0; i < attributes.size(); i++) {
            Attribute a = attributes.get(i);
            if (a.getNamespacePrefix().equals(prefix)) {
                if (a.getNamespaceURI().equals(URI)) break;
                else {
                    throw new NamespaceException(
                     "Attribute prefix conflicts with additional namespace");
                }
            }   
        }
        
        
        namespaces.put(prefix, URI);
        
    }

    /**
     * Removes the mapping of the speciifed prefix. This method only
     * removes additional namespaces added with 
     * <code>addNamespaceDeclaration</code>.
     * It has no effect on namespaces of elements and attributes.
     * 
     * @param prefix - the prefix whose declaration should be removed
     */
    public final void removeNamespaceDeclaration(String prefix) {
        namespaces.remove(prefix);
    }


    /**
     * Returns a read-only list of the additonal
     * namespaces declared on  this element.
     * The list is dead; that is, changes to the element are not reflected
     * in the contents of the list, and vice versa.
     * 
     * @return extra namespace declarations on this element
     */
    public final Namespaces getAdditionalNamespaces() {
        return namespaces.copy();
    }

    /**
     * Returns a string containing the XML serialization of this element.
     * This includes the element and all its attributes and descendants.
     * However, it may not contain namespace declarations for namespaces
     * inherited from ancestor elements.
     * 
     * 
     * @return the String form of this element
     * 
     * @see nu.xom.Node#getStringForm()
     * 
     * 
     */
    public final String getStringForm() {
    
        StringBuffer result = new StringBuffer();
        
        result.append("<");
        result.append(getQualifiedName());
        
        // Namespace
        if (!"".equals(URI)) {
            TreeNode parent = getParent();
            String parentURI = null;
            if (parent instanceof Element) {
                parentURI = ((Element) parent).getNamespaceURI(prefix);
            } 
            if (!URI.equals(parentURI)) {
                if (prefix.equals("")) {
                    result.append(" xmlns=\"");
                    result.append(URI);
                    result.append("\"");
                }
                else {
                    result.append(" xmlns:");
                    result.append(prefix);
                    result.append("=\"");
                    result.append(URI);
                    result.append("\"");                
                }
            }   
        }
        
        // attributes
        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = attributes.get(i);
            result.append(' ');
            result.append(attribute.getStringForm());   
            if (!"".equals(attribute.getNamespaceURI())) {
                String parentURI = this.getNamespaceURI(attribute.getNamespacePrefix());
                if (!attribute.getNamespaceURI().equals(parentURI)) {
                    result.append(" xmlns:" + attribute.getNamespacePrefix() + "=\"");
                    result.append(URI);
                    result.append("\"");
                }
            }   
        }       
        
        // additional namespaces
        for (int i = 0; i < namespaces.size(); i++) {
            String prefix = namespaces.getPrefix(i);
            String uri = namespaces.getURI(prefix);
            result.append(" xmlns:");   
            result.append(prefix);  
            result.append("=\"");   
            result.append(uri); 
            result.append('\"');
        }
        
        result.append('>');
        // children
        for (int i = 0; i < getChildCount(); i++) {
            result.append(getChild(i).getStringForm()); 
        }
        
        result.append("</");
        result.append(getQualifiedName());
        result.append(">");
        
        return result.toString();
        
    }

    /**
     * 
     * 
     * @see nu.xom.Node#getValue()
     */
    public final String getValue() {

        StringBuffer result = new StringBuffer();
        for (int i = 0; i < getChildCount(); i++) {
           Node child = getChild(i);
           if (child instanceof Text || child instanceof Element) {
               result.append(child.getValue());
           }
        }
        return result.toString();

    }

    /**
     * Creates a deep copy of this element with no parent,
     * that can be added to this document or a different one.
     * 
     * @see nu.xom.TreeNode#copy()
     */
    public final TreeNode copy() {
        return new Element(this);
    }

    /**
     * Returns a string representation of this element suitable for
     * debugging and diagnosis. This is <em>not</em>
     * the XML representation of the element.
     * 
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        return "[" + getClass().getName() + ": \"" + getQualifiedName() + "\"]";
    }

}