// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 *
 * <p>
 *  This is the generic superclass for all the contents of an XML document.
 *  There are exactly six kinds of nodes in XOM:
 * </p>
 * 
 * <ul>
 *   <li>Element</li>
 *   <li>Document</li>
 *   <li>Text</li>
 *   <li>Comment</li>
 *   <li>Attribute</li>
 *   <li>ProcessingInstruction</li>
 * </ul>
 * 
 * <p>
 *   Every instance of Node is an instance of one of these six classes
 *   (including, possibly, one of its subclasses).
 * </p>
 * 
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
public abstract class Node {
    
    private TreeNode parent = null;

   
    /**
     * Creates a new Node object.
     * Can only be invoked from subclasses.
     * 
     */
    public Node() {}

    /**
     * Returns the actual XML form of this node, such as might be
     * copied and pasted from the original document.
     * 
     * @return  A String containing an XML representation of this node
     */
    // Is this really the right name? What might be a better name????
    public abstract String getStringForm();

    /**
     * Returns the XPath 1.0 string-value of this node.
     * 
     * @return  the XPath 1.0 string-value of this node
     */
    public abstract String getValue();
    
    /**
     * 
     * Returns the document that contains this node,
     * or null if this node is not currently part of a document.
     * Each node belongs to no more than one document at a time.
     * If this node is a <code>Document</code>, then it returns this node.
     *
     * @return the <code>Document</code> which this node is a part of 
     */
    public final Document getDocument() {
        Node parent = this;
        while (parent != null && !(parent instanceof Document)) {
            parent = parent.getParent();
        }
        return (Document) parent;
    }
    
    /**
     * 
     * Returns the node that contains this node,
     * or null if this node does not have a parent.
     *
     * @return the parent <code>Element</code> or <code>Document</code> that contains this node 
     */
    public final TreeNode getParent() {
        return this.parent; 
    }
    
    
    final void setParent(TreeNode parent) {
        this.parent = parent;   
    }
    
    /**
     * Removes this node from its parent so that it can be added 
     * to a different parent node or document.
     */
    public final void detach() {

        if (parent == null) return;
        if (this instanceof TreeNode) {
            parent.removeChild((TreeNode) this);
        }
        else if (this instanceof Attribute) {
            Element element = (Element) parent;
            element.removeAttribute((Attribute) this);
        }

    }
    
    /**
     * Tests for node identity. That is, two 
     * <code>Node</code> objects are equal
     * if and only if they are the same object.
     * 
     * @see java.lang.Object#equals(Object)
     */
    public final boolean equals(Object o) {
        return super.equals(o); 
    }       

    /**
     * Returns a unique identifier for this node.
     * The value returned is the same as returned by super.hashCode()
     * because nodes use identity semantics. 
     * 
     * @see java.lang.Object#hashCode()
     */
    public final int hashCode() {
        return super.hashCode();    
    }
    
    /**
     * Returns a string representation of this node suitable for
     * debugging and diagnosis. This is <em>not</em>
     * the XML representation of this node.
     * 
     * @see java.lang.Object#toString()
     */
    public abstract String toString();
            
}
