// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 * <p>
 * Subclassing this class allows builders to produce
 * instance of subclasses (for example,
 * <code>HTMLElement</code>) instead of the
 * base classes.
 * </p>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 * 
 */
public class NodeFactory {


    /**
     * Creates a new <code>Element</code> in no namespace with
     * the specified name.
     * 
     * @param name the local name of the element
     * 
     * @return the new Element
     */
    public Element makeElement(String name) {
        return new Element(name);   
    }

    /**
     * Creates a new <code>Element</code> in the specified namespace with
     * the specified name.
     * 
     * @param name the qualified name of the element
     * @param namespace the namespace URI
     * 
     * @return the new Element
     */
    public Element makeElement(String name, String namespace) {
        return new Element(name, namespace);    
    }

    /**
     * Creates a new <code>Attribute</code> in no namespace with
     * the specified name.
     * 
     * @param name the local name of the attribute
     * @param value the attribute value
     * 
     * @return the new Attribute
     */
    public Attribute makeAttribute(String name, String value) {
        return new Attribute(name, "", value, Attribute.Type.UNDECLARED);
    }

    /**
     * Creates a new <code>Attribute</code> in no namespace with
     * the specified name and type.
     * 
     * @param name the local name of the attribute
     * @param value the attribute value
     * @param type the attribute type
     * 
     * @return the new Attribute
     */
    public Attribute makeAttribute(String name, String value, Attribute.Type type) {
        return new Attribute(name, "", value, type);
    }

    /**
     * Creates a new <code>Attribute</code> in the specified namespace with
     * the specified name and type.
     * 
     * @param name the local name of the attribute
     * @param URI the namespace URI
     * @param value the attribute value
     * @param type the attribute type
     * 
     * @return the new Attribute
     */
    public Attribute makeAttribute(String name, String URI, String value, Attribute.Type type) {
        return new Attribute(name, URI, value, type);
    }

    /**
     * Creates a new <code>Comment</code>.
     * 
     * @param data the complete text content of the comment
     * 
     * @return the new Comment
     */
    public Comment makeComment(String data) {
        return new Comment(data);   
    }

    /**
     * Creates a new <code>DocType</code> with a root element name,
     * a system ID, and a public ID.
     * 
     * @param rootElementName the declared, qualified name for the root element
     * @param publicID the public ID of the external DTD subset
     * @param systemID the URL of the external DTD subset
     * 
     * @return the new DocType
     */
    public DocType makeDocType(String rootElementName, String publicID, String systemID) {
        return new DocType(rootElementName, publicID, systemID);    
    }

    /**
     * Creates a new <code>DocType</code> with a root element name,
     * a system ID, and a null public ID.
     * 
     * @param rootElementName the declared, qualified name for the root element
     * @param systemID the URL of the external DTD subset
     * 
     * @return the new DocType
     */
    public DocType makeDocType(String rootElementName, String systemID) {
        return makeDocType(rootElementName, null, systemID);    
    }

    /**
     * Creates a new <code>DocType</code> with a root element name
     * but no public or system ID.
     * 
     * @param rootElementName the declared, qualified name for the root element
     * 
     * @return the new DocType
     */
    public DocType makeDocType(String rootElementName) {
        return makeDocType(rootElementName, null, null);    
    }

    /**
     * Creates a new <code>Document</code> with 
     * the specified root element.
     * 
     * @param root the root Element
     * 
     * @return the newly created document
     */
    public Document makeDocument(Element root) {
        return new Document(root);  
    }

    /**
     * Creates a new <code>Text</code> node.
     * 
     * @param data the complete text content of the node
     * 
     * @return the new Text
     */
    public Text makeText(String data) {
        return new Text(data);  
    }

    /**
     * Creates a new <code>ProcessingInstruction</code> with
     * the specified target and data.
     * 
     * @param target the target of the processing instruction
     * @param data the data of the processing instruction
     * 
     * @return the new ProcessingInstruction
     */
    public ProcessingInstruction makeProcessingInstruction(String target, String data) {
        return new ProcessingInstruction(target, data); 
    }

}
