// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 * <p>
 *  Represents an XML processing instruction.
 *  Each processing instruction has two key properties:
 * </p>
 * 
 * <ul>
 *   <li>The target, a non-colonized name</li>
 *   <li>The data, a string which does not contain the two character
 *       sequence <code>?&gt;</code>. The syntax of the data
 *       depends completely on the processing instruction. 
 *       Other than forbidding <code>?&gt;</code>, XML defines
 *       no rules for processing instruction data.
 *   </li>
 * </ul>
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
public class ProcessingInstruction extends LeafNode {

    private String target;
    private String data;

    /**
     * Create a processing instruction with a certain target and data.
     * 
     * @param target the target of the processing instruction
     * @param data the processing instruction data
     * 
     * @throws IllegalTargetException - if the target is not a non-colonized name
     *                                  or is made of the string "xml" in any case.
     * @throws IllegalDataException if data contains "?>" or any other illegal characters
     */
    public ProcessingInstruction(String target, String data) {
        setTarget(target);  
        setData(data);  
    }

    /**
     * Create a copy of a processing instruction.
     * 
     * @param instruction the processing instruction to copy
     * 
     */
    public ProcessingInstruction(ProcessingInstruction instruction) {
        this(instruction.target, instruction.data);  
    }

    /**
     * Returns the processing instruction target.
     * 
     * @return the target
     */
    public final String getTarget() {
        return target;
    }

    /**
     * Sets the target.
     * 
     * @param target The new target
     * 
     * @throws IllegalTargetException if the proposed target is not an XML 1.0 non-colonized name
     * @throws XMLException if the proposed target does not satisfy the local constraints
     */
    public final void setTarget(String target) {
        Verifier.checkProcessingInstructionTarget(target);
        checkTarget(target);
        this.target = target;
    }

    /**
     * Subclasses can override this method to perform additional checks
     * on the data beyond what XML 1.0 requires.
     * For example, an XMLStylesheet subclass might throw an exception
     * if the target were not xml-stylesheet.
     * 
     * @param target The potential target to check.
     * 
     * @throws XMLException if the proposed target does not satisfy the local constraints
     */
    protected void checkTarget(String target) {}

    /**
     * Sets the data.
     * 
     * @param data The data to set
     * 
     * @throws XMLException if the proposed data does not satisfy the local constraints
     */
    public final void setData(String data) {
        Verifier.checkProcessingInstructionData(data);
        checkData(data);
        this.data = data;
    }

    /**
     * Returns the processing instruction data.
     * 
     *  @see nu.xom.Node#getValue()
     * 
     * @return  the data of the processing instruction
     */
    public final String getValue() {
        return data;
    }

    /**
     * Subclasses can override this method to perform additional checks
     * on the data beyond what XML 1.0 requires.
     * For example, an <code>XMLStylesheet</code> subclass might throw an exception
     * if the data did not have a pseudo-attribute format.
     * 
     * @param data The potential data to check.
     * 
     * @throws XMLException if the proposed data does not satisfy the local constraints
     */
    protected void checkData(String data) {}

    /**
     * Returns the actual XML form of this processing instruction, such as might be
     * copied and pasted from the original document.
     * 
     * @return  An XML representation of this processing instruction as a <code>String</code>
     */
    public final String getStringForm() {
        return "<?" + target + " " + data + "?>";   
    }

    /**
     * Returns a deep copy of this processing instruction with no parent,
     * that can be added to this document or a different one.
     * 
     * @see nu.xom.TreeNode#copy()
     */
    public final TreeNode copy() {
        return new ProcessingInstruction(target, data);
    }

    /**
     * Returns a <code>String</code> representation of this processing instruction suitable for
     * debugging and diagnosis. This is <em>not</em>
     * the XML representation of this processing instruction.
     * 
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        return "[" + getClass().getName() + ": target=" 
         + target + "; data=" + data +"]";
    }

}
