// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;

/**
 * 
 * <p>
 * This class walks a XOM tree, while firing events into a
 * SAX2 <code>ContentHandler</code>
 * </p>
 * 
 * Needs to be tested????
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 * 
 */
public class SAXConverter {

    private ContentHandler contentHandler;
/*    private DeclHandler    declHandler;
    private DTDHandler     dtdHandler; */
    private LexicalHandler lexicalHandler;

    /**
     * Creates a new <code>SAXConverter</code>
     * 
     * @param handler SAX2 <code>ContentHandler</code> that receives the data
     * 
     * @thorws NullPointerException if handler is null
     * 
     */
    public SAXConverter(ContentHandler handler) {
        setContentHandler(handler);
    }

/* Hold for when we add notations and entities
    public void setDTDHandler(DTDHandler handler) {
        this.dtdHandler = handler;
    }

    public DTDHandler getDTDHandler() {
        return this.dtdHandler;
    }*/

    /**
     * Set the <code>ContentHandler</code> for this converter.
     * 
     * @param handler SAX2 ContentHandler that receives the data
     * 
     * @thorws NullPointerException if handler is null
     * 
     */
    public void setContentHandler(ContentHandler handler) {
        if (handler == null) throw new NullPointerException("ContentHandler must be non-null.");
        this.contentHandler = handler;
    }

    /**
     * Returns the <code>ContentHandler</code>.
     * 
     * @return ContentHandler
     */
    public ContentHandler getContentHandler() {
        return this.contentHandler;
    }

/*
    public void setDeclHandler(DeclHandler handler) {
        this.declHandler = handler;
    }

    public DeclHandler getDeclHandler() {
        return this.declHandler;
    } */

    /**
     * Sets the optional <code>LexicalHandler</code> for this converter.
     * The only lexical events the converter supplies are comments.
     * 
     * @param handler the LexicalHandler; may be null to turn off lexical events
     */
    public void setLexicalHandler(LexicalHandler handler) {
        this.lexicalHandler = handler;
    }

    /**
     * Returns the <code>LexicalHandler</code> for this
     * converter.
     * 
     * @return LexicalHandler
     */
    public LexicalHandler getLexicalHandler() {
        return this.lexicalHandler;
    }
    
    /**
     * Feed a document through this converter.
     * 
     * @param doc the document to pass to SAX
     * @throws SAXException if the <code>ContentHandler</code> or <code>LexicalHandler</code> 
     *      throws an exception
     */
    public void convert(Document doc) throws SAXException {
        
        contentHandler.startDocument();        
        for (int i = 0; i < doc.getChildCount(); i++) {
            process(doc.getChild(i));
        }        
        contentHandler.endDocument();
        
    }
    
    private void process(TreeNode node) throws SAXException {
        
        if (node instanceof Element) {
            Element element = (Element) node;
            // start prefix mapping
            Namespaces additional = element.getAdditionalNamespaces();
            for (int i = 0; i < additional.size(); i++) {
                String prefix = additional.getPrefix(i);
                contentHandler.startPrefixMapping(prefix , additional.getURI(prefix));  
            }
            TreeNode parentNode = element.getParent();
            Attributes attributes = element.getAttributes();
            if (parentNode instanceof Element) {
                Element parent = (Element) parentNode;
                // now handle element's prefix if not declared on ancestor
                String prefix = element.getNamespacePrefix();
                if (!element.getNamespaceURI(prefix).equals(parent.getNamespaceURI(prefix))) {
                    contentHandler.startPrefixMapping(prefix, element.getNamespaceURI(prefix));  
                }
                
                // Handle attributes' prefixes if not declared on ancestor
                for (int i = 0; i < attributes.size(); i++) {
                    Attribute att = attributes.get(i);
                    String attPrefix = att.getNamespacePrefix();
                    if (!element.getNamespaceURI(attPrefix).equals(parent.getNamespaceURI(attPrefix))
                      && !element.getNamespacePrefix().equals(attPrefix)) {
                        contentHandler.startPrefixMapping(attPrefix, element.getNamespaceURI(attPrefix));  
                    }
                }                
            }
            else { // declare all prefixes
                String prefix = element.getNamespacePrefix();
                if (!prefix.equals("")) {
                    contentHandler.startPrefixMapping(prefix , element.getNamespaceURI());  
                }
                
                // Handle attributes' prefixes if not declared on ancestor
                for (int i = 0; i < attributes.size(); i++) {
                    Attribute att = attributes.get(i);
                    String attPrefix = att.getNamespacePrefix();
                    if (!attPrefix.equals("") && !attPrefix.equals(element.getNamespacePrefix())) {
                        contentHandler.startPrefixMapping(attPrefix, att.getNamespaceURI());  
                    }
                }
                
            }
            
            
            // add attributes
            AttributesImpl saxAttributes = new AttributesImpl();
            for (int i = 0; i < attributes.size(); i++) {
                Attribute attribute = attributes.get(i);
                saxAttributes.addAttribute(attribute.getNamespaceURI(),
                  attribute.getLocalName(),
                  attribute.getQualifiedName(),
                  getSAXType(attribute),
                  attribute.getValue());
            }

            
            contentHandler.startElement(element.getNamespaceURI(),
              element.getLocalName(), element.getQualifiedName(), saxAttributes);
            for (int i = 0; i < element.getChildCount(); i++) {
                process(element.getChild(i));   
            }
            contentHandler.endElement(element.getNamespaceURI(),
              element.getLocalName(), element.getQualifiedName());
            
            // end prefix mappings
            for (int i = 0; i < additional.size(); i++) {
                String prefix = additional.getPrefix(i);
                contentHandler.endPrefixMapping(prefix);  
            }
            if (parentNode instanceof Element) {
                Element parent = (Element) parentNode;
                // now handle element's prefix if not declared on ancestor
                String prefix = element.getNamespacePrefix();
                if (!element.getNamespaceURI(prefix).equals(parent.getNamespaceURI(prefix))) {
                    contentHandler.endPrefixMapping(prefix);  
                }
                
                // Handle attributes' prefixes if not declared on ancestor
                for (int i = 0; i < attributes.size(); i++) {
                    Attribute att = attributes.get(i);
                    String attPrefix = att.getNamespacePrefix();
                    if (!element.getNamespaceURI(attPrefix).equals(parent.getNamespaceURI(attPrefix))
                      && !element.getNamespacePrefix().equals(attPrefix)) {
                        contentHandler.endPrefixMapping(attPrefix);  
                    }
                }                
            }
            else { // undeclare all prefixes
                String prefix = element.getNamespacePrefix();
                if (!prefix.equals("")) {
                    contentHandler.endPrefixMapping(prefix);  
                }
                
                // Handle attributes' prefixes if not declared on ancestor
                for (int i = 0; i < attributes.size(); i++) {
                    Attribute att = attributes.get(i);
                    String attPrefix = att.getNamespacePrefix();
                    if (!attPrefix.equals("") && !attPrefix.equals(element.getNamespacePrefix())) {
                        contentHandler.endPrefixMapping(attPrefix);  
                    }
                }
                
            }
            
        }
        else if (node instanceof Text) {
            String data = node.getValue();
            contentHandler.characters(data.toCharArray(), 0, data.length());
        }
        else if (node instanceof ProcessingInstruction) {
            ProcessingInstruction instruction = (ProcessingInstruction) node;
            contentHandler.processingInstruction(instruction.getTarget(), instruction.getValue());
        }
        else if (node instanceof Comment && lexicalHandler != null) {
            String data = node.getValue();
            lexicalHandler.comment(data.toCharArray(), 0, data.length());            
        }
        else if (node instanceof DocType && lexicalHandler != null) {
            DocType type = (DocType) node;
            lexicalHandler.startDTD(type.getRootElementName(), 
              type.getPublicID(), type.getSystemID());              
            lexicalHandler.endDTD();          
        }
        
    }
    
    private static String getSAXType(Attribute attribute) {

        Attribute.Type type = attribute.getType();
        if (type == Attribute.Type.UNDECLARED) return "CDATA";
        if (type == Attribute.Type.CDATA) return "CDATA";
        if (type == Attribute.Type.ID) return "ID";
        if (type == Attribute.Type.IDREF) return "IDREF";
        if (type == Attribute.Type.IDREFS) return "IDREFS";
        if (type == Attribute.Type.NMTOKEN) return "NMTOKEN";
        if (type == Attribute.Type.NMTOKENS) return "NMTOKENS";
        if (type == Attribute.Type.ENTITY) return "ENTITY";
        if (type == Attribute.Type.ENTITIES) return "ENTITIES";
        if (type == Attribute.Type.NOTATION) return "NOTATION";
        
        return "CDATA";
    }

}
