// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.BufferedWriter;
import java.io.Writer;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

/**
 * <p>
 *  A serializer allows you to output a <code>Document</code> object
 *  in a specific encoding
 *  using various options for controlling white space, indenting, and 
 *  line breaking. However, in general these do affect  the
 *  document's infoset. In particular, if you set either the maximum line length 
 *  or the indent size to a positive value, then the serializer will
 *  not respect input white space. It may trim leading and trailing space,
 *  condense runs of white space to a single space, convert carriage returns 
 *  and line feeds to spaces, add extra space where none was present before,
 *  and otherwise muck with the document's white space.
 *  The <code>xml:space</code> attribute is not respected. 
 *  The defaults, however, preserve all significant white space including
 *  ignorable white space. 
 * </p>
 * 
 * <p>
 *   Open question: should the OutputStream be passed to the individual write()
 *   methods rather than being a fundamental part of the serializer????
 * </p>
 * 
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
public class Serializer {

    private TextWriter escaper;

    /**
     * Create a new serializer that uses the UTF-8 encoding.
     * 
     * @param out The output stream to write the document on
     * 
     * @throws NullPointerException if out is null
     */
    public Serializer(OutputStream out) {
        if (out == null) throw new NullPointerException("Null OutputStream");
        try {
            Writer writer = new OutputStreamWriter(out, "UTF-8");
            writer = new BufferedWriter(writer);
            escaper = TextWriterFactory.getTextWriter(writer, "UTF-8"); 
        }
        catch (UnsupportedEncodingException ex) {
            throw new XMLException("The VM is broken. It does not understand UTF-8.");
        }
    }
    
    /**
     * <p>
     * Create a new serializer that uses a specified encoding.
     * The encoding must be recognized by the Java virtual machine.
     * Currently the following encodings are recognized by XOM:
     * </p>
     * 
     * <ul>
     *   <li>UTF-8</li>
     *   <li>UTF-16</li>
     *   <li>UCS2</li>
     *   <li>UCS4</li>
     *   <li>UnicodeBig</li>
     *   <li>UnicodeLittle</li>
     *   <li>ISO-8859-1</li>
     * </ul>
     * 
     * <p>
     *   More will be added in the future.
     *   You can use encodings not in this list as long as the local virtual
     *   machine supports them. However, characters may unnecessarily be output
     *   as character references.
     * </p>
     * 
     * @param out The output stream to write the document on
     * @param encoding The character encoding for the serialization
     * 
     * @throws NullPointerException if out or encoding is null
     * @throws UnsupportedEncodingException - if the VM does not support the requested encoding
     *  
     */
    public Serializer(OutputStream out, String encoding)
     throws UnsupportedEncodingException {
        if (out == null) throw new NullPointerException("Null OutputStream");
        if (encoding == null) throw new NullPointerException("Null encoding");
        Writer writer = new OutputStreamWriter(out, encoding);
        writer = new BufferedWriter(writer);
        escaper = TextWriterFactory.getTextWriter(writer, encoding);    
    }

    
    /**
     * Serializes a document onto the output stream using the current options.
     * 
     * @param doc the <code>Document</code> to serialize.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void write(Document doc) throws IOException {
        escaper.reset();
        escaper.writeMarkup("<?xml version=\"1.0\" encoding=\"");
        escaper.writeMarkup(escaper.getEncoding());
        escaper.writeMarkup("\"?>");
        escaper.breakLine();
        for (int i = 0; i < doc.getChildCount(); i++) {
            write(doc.getChild(i)); 
            escaper.breakLine();
        }       
        escaper.flush();
    }
   
    
    /**
     * Serializes an element onto the output stream using the current options.
     * The result is guaranteed to be well-formed. If <code>element</code>
     * does not have a parent element,
     * it will also be namespace well-formed.
     * 
     * 
     * @param element the <code>Element</code> to serialize.
     * 
     * @throws IOException if the underlying <code>OutputStream</code> throws an I/O error
     */
    public void write(Element element) throws IOException {
        
        if (escaper.isIndenting()) escaper.breakLine();
        escaper.writeMarkup("<");
        escaper.writeMarkup(element.getQualifiedName());
        
        // Namespace
        String URI = element.getNamespaceURI();
        String prefix = element.getNamespacePrefix();
        if (!"".equals(URI)) {
            TreeNode parent = element.getParent();
            String parentURI = null;
            if (parent instanceof Element) {
                parentURI = ((Element) parent).getNamespaceURI(prefix);
            } 
            if (!URI.equals(parentURI)) {
                if (prefix.equals("")) {
                    escaper.writeMarkup(" xmlns=\"");
                    escaper.writePCDATA(URI);
                    escaper.writeMarkup("\"");
                }
                else {
                    escaper.writeMarkup(" xmlns:");
                    escaper.writeMarkup(prefix);
                    escaper.writeMarkup("=\"");
                    escaper.writePCDATA(URI);
                    escaper.writeMarkup("\"");              
                }
            }   
        }
        
        // attributes
        Attributes attributes = element.getAttributes();
        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = attributes.get(i);
            escaper.writeMarkup(' ');
            escaper.writeMarkup(attribute.getQualifiedName());
            escaper.writeMarkup("=\"");
            escaper.writeAttributeValue(attribute.getValue());
            escaper.writeMarkup("\"");
            String attributeURI = attribute.getNamespaceURI();
            if (!"".equals(attributeURI)) {
                Element parent = (Element) attribute.getParent();
                String parentURI = parent.getNamespaceURI(attribute.getNamespacePrefix());
                if (!attributeURI.equals(parentURI)) {
                    escaper.writeMarkup(" xmlns:");
                    escaper.writeMarkup(attribute.getNamespacePrefix());
                    escaper.writeMarkup("=\"");
                    escaper.writePCDATA(attributeURI);
                    escaper.writeMarkup("\"");
                }
            }   
        }       
        
        // additional namespaces
        Namespaces namespaces = element.getAdditionalNamespaces();
        for (int i = 0; i < namespaces.size(); i++) {
            String additionalPrefix = namespaces.getPrefix(i);
            String uri = namespaces.getURI(additionalPrefix);
            escaper.writeMarkup(" xmlns:"); 
            escaper.writeMarkup(additionalPrefix);  
            escaper.writeMarkup("=\""); 
            escaper.writePCDATA(uri);   
            escaper.writeMarkup('\"');
        } 
        
        if (element.hasChildren()) {
            escaper.writeMarkup('>');
            escaper.incrementIndent();
            // children
            for (int i = 0; i < element.getChildCount(); i++) {
                write(element.getChild(i)); 
            }
            escaper.decrementIndent();
            if (escaper.getIndent() > 0) {
                Node firstChild = element.getFirstChild();
                if (!(firstChild instanceof Text) 
                  || firstChild.getValue().trim().equals("")) escaper.breakLine();
            }
            escaper.writeMarkup("</");
            escaper.writeMarkup(element.getQualifiedName());
            escaper.writeMarkup(">");
        }
        else {
            escaper.writeMarkup("/>");   
        }
        escaper.flush();
        
    }
    
    /**
     * Serializes a <code>Comment</code> object
     * onto the output stream using the current options.
     * 
     * @param comment the <code>Comment</code> to serialize.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void write(Comment comment) throws IOException {
        if (escaper.isIndenting()) escaper.breakLine();
        escaper.writeMarkup("<!--");
        escaper.writeMarkup(comment.getValue());
        escaper.writeMarkup("-->");
    }
    
    /**
     * Serializes a <code>ProcessingInstruction</code> object
     * onto the output stream using the current options.
     * 
     * @param instruction the <code>ProcessingInstruction</code> to serialize.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void write(ProcessingInstruction instruction) throws IOException {
        if (escaper.isIndenting()) escaper.breakLine();
        escaper.writeMarkup("<?");
        escaper.writeMarkup(instruction.getTarget());
        escaper.writeMarkup(' ');
        escaper.writeMarkup(instruction.getValue());
        escaper.writeMarkup("?>");      
    }
    
    /**
     * <p>
     * Serializes a <code>Text</code> object
     * onto the output stream using the current options.
     * Reserved characters such as &lt;, &gt; and "
     * are escaped using the standard entity references such as
     * <code>&amp;lt;</code>, <code>&amp;gt;</code>, and <code>&amp;quot;</code>.
     * </p>
     * 
     * <p>
     *   Characters which cannot be encoded in the current character set
     *   (for example, &Omega; in ISO-8859-1) are encoded using 
     *   character references.
     * </p>
     * 
     * <p>
     *   Unsupported character sets encode all non-ASCII characters.
     *   Supported character sets currently include:
     * </p>
     * 
     * <ul>
     *   <li>Unicode in its various encodings</li>
     *   <li>Latin-1 (ISO-8859-1)</li>
     * </ul>
     * 
     * <p>
     *   Non-ASCII characters from other character sets will probably be hexadecimally escaped.
     *   even when they don't need to be. More standard character sets will be added in the future.
     *   This will not require any changes to the public API.
     * </p> 
     * 
     * @param text the <code>Text</code> to serialize.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void write(Text text) throws IOException {
        escaper.writePCDATA(text.getValue());
    }   

    /**
     * Serializes a node
     * onto the output stream using the current options.
     * 
     * @param node the <code>TreeNode</code> to serialize.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void write(TreeNode node) throws IOException {
        
        if (node instanceof Text) {
            write((Text) node);
        }
        else if (node instanceof Element) {
            write((Element) node);
        }
        else if (node instanceof Comment) {
            write((Comment) node);
        }
        else if (node instanceof ProcessingInstruction) {
            write((ProcessingInstruction) node);
        }
        else if (node instanceof Document) {
            write((Document) node);
        }
        else if (node instanceof DocType) {
            write((DocType) node);
        }
        else {
            throw new XMLException("Serializer cannot directly serialize a " + node.getClass());    
        }
        
    }
    
    /**
     * Flush the data onto the output stream.
     * It is not enough to flush the output stream.
     * You must flush the serializer object itself becuase it
     * uses some internal buffering.
     * 
     * @throws IOException  if the underlying <code>OutputStream</code> encounters an I/O error
     */
    public void flush() throws IOException {
        escaper.flush();    
    }

    /**
     * Returns the number of spaces this serializer indents.
     * 
     * @return the number of spaces this serializer indents
     */
    public int getIndent() {
        return escaper.getIndent();
    }


    /**
     * Sets the number of spaces to indent each 
     * successive level in the hierarchy.
     * Use 0 for no extra indenting. The maximum indentation is in
     * general limited to approximately half the maximum line length.
     * The serializer will not indent further than that no matter how many
     * levels deep the hierarchy is.
     * 
     * @param indent The indent to set
     */
    public void setIndent(int indent) {
        escaper.setIndent(indent);
    }

    /**
     * Returns the <code>String</code> used as a line separator.
     * This is always <code>"\n"</code>, <code>"\r"</code>, or <code>"\r\n"</code>.
     * 
     * @return the line separator
     */
    public String getLineSeparator() {
        return escaper.getLineSeparator();
    }

    /**
     * Sets the lineSeparator. This can only be one of the three
     * strings <code>"\n"</code>, <code>"\r"</code>, or <code>"\r\n"</code>.
     * All other values are forbidden.
     * Line separators in the input text are only adjusted if this
     * serializer is either indenting or wrapping.
     * 
     * @param lineSeparator The lineSeparator to set
     * 
     * @throws IllegalArgumentException if you attempt to use any line separator other
     *                                    than <code>"\n"</code>, <code>"\r"</code>, or <code>"\r\n"</code>.
     * 
     */
    public void setLineSeparator(String lineSeparator) {
        escaper.setLineSeparator(lineSeparator);  
    }

    /**
     * Returns the preferred maximum line length.
     * 
     * @return the maximum line length.
     */
    public int getMaxLength() {
        return escaper.getMaxLength();
    }

    /**
     * Sets the suggested maximum line length for this serializer.
     * Setting this to 0 indicates that no automatic wrapping is to be performed.
     * When a line approaches this length, the serializer begins looking
     * for opportunities to break the line. Generally it will break on 
     * any ASCII white space character (tab, carriage return, linefeed,
     * and space).
     * In some circumstances the serializer may not be able to break the line
     * before the maximum length is reached.
     * For instance, if an element name is longer than the maximum line length 
     * the only way to correctly serialize it is to exceed the maximum line length.
     * In this case the serializer will exceed the maximum line length.
     * 
     * @param maxLength The suggested maximum line length
     */
    public void setMaxLength(int maxLength) {
        escaper.setMaxLength(maxLength);
    }

}