// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

/**
 * <p>
 *   This class represents a run of text. The system endeavors to merge text
 *   nodes as necessary so all <code>Text</code> objects contain the maximum possible
 *   contiguous run of text. (not yet implemented????)
 *   CDATA sections are not treated differently than 
 *   normal text.
 * </p>

 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
public class Text extends LeafNode {

    private String data;

    /**
     * <p>
     * This constructor creates a new <code>Text</code> object. The data is checked for 
     * legality according to XML 1.0 rules. Characters that can be serialized
     * such as &lt; and &amp; are allowed. However, characters such as the 
     * form feed, null, vertical tab.
     * unmatched halves of surrogate pairs,
     * and 0xFFFE and 0xFFFF are not allowed.
     * </p>
     * 
     * @param data The initial text of the object
     *
     * @throws IllegalDataException if data contains any characters which are illegal
     *  in well-formed XML 1.0 such as null, vertical tab, or unmatched halves of surrogate pairs.
     */
    public Text(String data) {
        setData(data);
    }

    /**
     * <p>
     * Creates a copy of the specified <code>Text</code> object
     * </p>
     * 
     * @param text The <code>Text</code> object to copy
     */
    public Text(Text text) {
        this(text.data);
    }

    /**
     * <p>
     * This sets the content of the <code>Text</code> object to the specified data.
     * The data is checked for 
     * legality according to XML 1.0 rules. Characters that can be serialized
     * such as &lt; and &amp; are allowed. However, characters such as the 
     * form feed, null, vertical tab.
     * unmatched halves of surrogate pairs,
     * and 0xFFFE and 0xFFFF are not allowed.
     * </p>
     * 
     * @param data The text to install in the object
     * 
     * @throws IllegalDataException if data contains any characters which are illegal
     *  in well-formed XML 1.0 such as null, vertical tab, or unmatched halves of surrogate pairs.
     */
    public final void setData(String data) {
        Verifier.checkCharacterData(data);
        check(data);
        if (data == null) data = "";
        this.data = data;   
    }

    /**
     * <p>
     * Subclasses can override this method to perform additional checks.
     * For example, an <code>Integer</code> subclass could verify that
     * the text was a legal base-10 integer. However, this can only be used
     * to add checks, not remove them. All text in text nodes must be 
     * potentially well-formed when serialized. 
     * </p>
     * 
     * @param data The text to check
     */
    protected void check(String data) {}

    /**
     * <p>
     * Returns the XPath 1.0 string-value of this <code>Text</code> node.
     * The XPath string-value of a text node is the same as the text of the node.
     * </p>
     *  
     * @return The content of the node.
     * 
     * @see nu.xom.Node#getValue()
     */
    public final String getValue() {
        return data;
    }

    /**
     * <p>
     * Returns a deep copy of this <code>Text</code> with no parent,
     * that can be added to this document or a different one.
     * </p>
     * 
     * @see nu.xom.TreeNode#copy()
     */
    public final TreeNode copy() {
        return new Text(data);
    }

    // add a split method that breaks at a certain position
    // by inserting a new node????

    /**
     * <p>
     * Returns a string containing the XML serialization of this text node.
     * </p>
     * 
     * @return the String form of this text node
     * 
     * @see nu.xom.Node#getStringForm()
     * 
     * 
     */
    public final String getStringForm() {
        return escapeText(data);    
    }

    private static String escapeText(String s) {
        
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < s.length(); i++) {
            result.append(escapeChar(s.charAt(i)));
        }
        return result.toString();
    }
    
    private static String escapeChar(char c) {
        
        switch (c) {
            case '<':
                return "&lt;";
            case '>':
                return "&gt;";
            case '&':
                return "&amp;";
            default: 
                return String.valueOf(c);   
        }
        
    }
    


    /**
     * <p>
     * Returns a <code>String</code> 
     * representation of this <code>Text</code> suitable for
     * debugging and diagnosis. This is <em>not</em>
     * the XML representation of this <code>Text</code> node.
     * </p>
     * 
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        
        if (data.length() <= 40) {
            return "[" + getClass().getName() + ": " + data + "]";
        }
        
        return "[" + getClass().getName() + ": " + data.substring(35) + "...]";
    }

}