// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import java.io.IOException;
import java.io.Writer;


/**
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
abstract class TextWriter {

    protected Writer out;
    protected String encoding;
    
    private String lineSeparator = "\r\n";
    private int    maxLength = 0;
    private int    indent = 0;
    private String indentString = "";
    private int    column = 0;
    
    protected TextWriter(Writer out, String encoding) {
        this.out = out; 
        this.encoding = encoding;
    }
    
    public void reset() {
        column = 0; 
        fakeIndents = 0; 
        lastCharacterWasSpace = false; 
    }

    private boolean lastCharacterWasSpace = false;
    
    public void writePCDATA(char c) throws IOException {
        if (needsEscaping(c)) {
            String s = "&0x" + Integer.toHexString(c) + ';';
            out.write(s);
            column += s.length();
            lastCharacterWasSpace = false;
        }
        else if (c == '&') {
            out.write("&amp;");
            column += 5;
            lastCharacterWasSpace = false;
        }
        else if (c == '<') {
            out.write("&lt;");
            column += 4;
            lastCharacterWasSpace = false;            
        }
        else if (c == '>') {
            out.write("&gt;");
            column += 4;
            lastCharacterWasSpace = false;                        
        }
        else write(c); 
    }

    // This is the same as writePCDATA except that it
    // also needs to escape " as &quot;
    // I'm not escaping the single quote because we
    // always use double quotes to contain 
    // values
    public void writeAttributeValue(char c) throws IOException {
        if (needsEscaping(c)) {
            String s = "&0x" + Integer.toHexString(c) + ';';
            out.write(s);
            column += s.length();
            lastCharacterWasSpace = false;
        }
        else if (c == '"') {
            out.write("&quot;");
            column += 6;
            lastCharacterWasSpace = false;
        }
        else if (c == '&') {
            out.write("&amp;");
            column += 5;
            lastCharacterWasSpace = false;
        }
        else if (c == '<') {
            out.write("&lt;");
            column += 4;
            lastCharacterWasSpace = false;            
        }
        else if (c == '>') {
            out.write("&gt;");
            column += 4;
            lastCharacterWasSpace = false;                        
        }
        else write(c); 
    }

    private void write(char c) throws IOException {
      if ((c == ' ' || c == '\r' || c == '\n' || c == '\t')) {
            if (needsBreak()) {
                breakLine();
            }
            else if (indent <= 0 && maxLength <= 0) {
                // If we're neither indenting nor wrapping
                // we need to preserve all white space
                out.write(c);
                column++;   
            }
            else if (!lastCharacterWasSpace) {
                out.write(' ');
                column++;
            }
            lastCharacterWasSpace = true;
        }
        else {  
            out.write(c);
            column++;
            lastCharacterWasSpace = false;
        }            
    }


    private boolean needsBreak() {
        if (maxLength <= 0) return false;
        return column > maxLength-1;   
    }

    public void breakLine() throws IOException {
        out.write(lineSeparator);
        out.write(indentString);
        column = indentString.length();
        lastCharacterWasSpace = true;
    }
    
    public void writeMarkup(char c) throws IOException {
        if (needsEscaping(c)) {
            throw new XMLException("Cannot use the character &0x" + 
              Integer.toHexString(c) + "; in the " + getEncoding() + " encoding.");
        }
        write(c);   

    }

    public void writePCDATA(String s) throws IOException {
        for (int i=0; i < s.length(); i++) {
            writePCDATA(s.charAt(i));
        }   
    }

    public void writeAttributeValue(String s) throws IOException {
        for (int i=0; i < s.length(); i++) {
            writeAttributeValue(s.charAt(i));
        }   
    }

    public void writeMarkup(String s) throws IOException {
        for (int i=0; i < s.length(); i++) {
            writeMarkup(s.charAt(i));
        }   
    }
    
    public boolean isIndenting() {
        return indentString.length() > 0;   
    }


    private int fakeIndents = 0;
    
    public void incrementIndent() {
        
        StringBuffer newIndent = new StringBuffer(indentString);
        for (int i = 0; i < indent; i++) {
            newIndent.append(' ');
        }
        
        // limit maximum indent to half of max length
        if (maxLength > 0 && newIndent.length() > maxLength / 2) fakeIndents++; 
        else this.indentString = newIndent.toString();
        
    }
    
    public void decrementIndent() {
        if (fakeIndents > 0) fakeIndents --;        
        else {
            this.indentString = indentString.substring(0, indentString.length()-indent);
        }
    }


    public String getEncoding() {
        return this.encoding;   
    }

    /**
     * Returns the String used as a line separator.
     * This is always "\n", "\r", or "\r\n".
     * 
     * @return the line separator
     */
    public String getLineSeparator() {
        return lineSeparator;
    }

    /**
     * Sets the lineSeparator. This can only be one of the three
     * strings "\n", "\r", or "\r\n".
     * All other values are forbidden.
     * 
     * @param lineSeparator The lineSeparator to set
     * 
     * @throws IllegalArgumentException if you attempt to use any line separator other
     *                                    than "\n", "\r", or "\r\n".
     * 
     */
    public void setLineSeparator(String lineSeparator) {
        if (lineSeparator.equals("\n") 
          || lineSeparator.equals("\r")
          || lineSeparator.equals("\r\n")) { 
            this.lineSeparator = lineSeparator;
        }
        throw new IllegalArgumentException("Illegal Line Separator");  
    }

    public String getIndentString() {
        return indentString;
    }

    /**
     * Returns the number of spaces this serializer indents.
     * 
     * @return the number of spaces this serializer indents
     */
    public int getIndent() {
        return indent;
    }


    /**
     * Returns the maximum line length.
     * 
     * @return the maximum line length.
     */
    public int getMaxLength() {
        return maxLength;
    }

    /**
     * Sets the suggested maximum line length for this serializer.
     * In some circumstances this may not be respected.
     * Not yet implemented.????
     * 
     * @param maxLength The maxLength to set
     */
    public void setMaxLength(int maxLength) {
        if (maxLength < 0) maxLength = 0;
        this.maxLength = maxLength;
    }

   /**
     * Sets the number of spaces to indent each 
     * successive level in the hierarchy.
     * Use 0 for no extra indenting.
     * 
     * @param indent The indent to set
     */
    public void setIndent(int indent) {
        this.indent = indent;
    }

    
    /**
     * Method flush.
     * @throws IOException
     */
    public void flush() throws IOException {
        out.flush();    
    }

    public abstract boolean needsEscaping(char c);

}
