// Copyright © 2002 Elliotte Rusty Harold
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of version 2.1 of 
// the GNU Lesser General Public License as published by the 
// Free Software Foundation.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You can contact Elliotte Rusty Harold by sending e-mail to
// elharo@metalab.unc.edu. Please include the word "XOM" in the
// subject line. The XOM home page is temporarily located at
// http://www.cafeconleche.org/XOM/  but will eventually move
// to http://www.xom.nu/

package nu.xom;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;


/**
 * 
 * This is the generic superclass for all the nodes which appear in 
 * the XML tree. It provides methods for navigation.
 * 
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
public abstract class TreeNode extends Node {

    private List children = new ArrayList();

    /**
     * This method returns true if and only if this node has child nodes.
     * This is not the same question as whether this node contains child <i>elements</i>.
     * This method may return true even if this node contains only comments, text nodes,
     * and processing instructions.
     * 
     * @return true if this node has children, false otherwise
     */
    public boolean hasChildren() {
        return !children.isEmpty(); 
    }


    /**
     * Returns the number of child nodes this node has.
     * This is always greater than or equal to 0.
     * 
     * @return the number of children of this node
     */
    public int getChildCount() {
        return children.size(); 
    }

    /**
     * <p>
     * Inserts a child node at the specified position.
     * Inserting at position 0 makes the child the first child
     * of this node. Inserting at the position <code>howManyChildren()</code>
     * makes the child the last child of the node.
     * </p>
     * 
     * <p>
     * All the other methods that add a node to the tree,
     * invoke this method ultimately.
     * </p>
     * 
     * @param position where to insert the child
     * @param child the node to insert
     * 
     * @throws IllegalAddException if this node cannot have child of
     *                               passed type
     * @throws MultipleParentException if <code>child</code> already has a parent
     * @throws NullPointerException if <code>child</code> is null
     * @throws IndexOutOfBoundsException if the position is negative or greater than
     *                                     the number of children of this node.
     */
    public void insertChild(int position, TreeNode child) {
        
        if (child == null) {
            throw new NullPointerException(
             "Tried to insert a null child in the tree");
        }
        if (child.getParent() != null) {
            throw new MultipleParentException("Child already has a parent.");
        }
        
        children.add(position, child);
        child.setParent(this);

    }

    /**
     * <p>
     * Converts a string to a text node and inserts that
     * node at the specified position.
     * </p>
     * 
     * @param position where to insert the child
     * @param child the node to insert
     * 
     * @throws IllegalAddException if this node cannot have a text child
     * @throws IndexOutOfBoundsException if the position is negative or greater than
     *                                     the number of children of the node.
     */
    public final void insertChild(int position, String text) {
       if (text == null) return;
       if (text.equals("")) return;
       insertChild(position, new Text(text));
    }

    /**
     * Appends a node to the children of this node.
     * 
     * @param child node to append to this node
     * 
     * @throws IllegalAddException if this node cannot have children of this type
     * @throws MultipleParentException if child already has a parent
     * 
     */
    public final void appendChild(TreeNode child) {
        insertChild(children.size(), child);
    }


    /**
     * Converts a string to a text node
     * and appends that node to the children of this node.
     * 
     * @param text String to add to this node.
     * 
     * @throws IllegalAddException if this node cannot have children of this type
     */
    public final void appendChild(String text) {
        insertChild(children.size(), new Text(text));
    }

    /**
     * Inserts a new child immediately before an existing child.
     * If <code>refChild</code> is not a child of this node, then
     * a <code>NoSuchChildException</code> is thrown. 
     * 
     * @param refChild the node before which the new node will be inserted
     * @param newChild the node inserted into the tree
     * 
     * @throws NoSuchChildException if <code>refChild</code> is not a child of this node
     * @throws NullPointerException if <code>newChild</code> is null
     * @throws MultipleParentException if <code>newChild</code> already has a parent
     * @throws IllegalAddException if this node cannot have children of this type
     */
    public final void insertBefore(TreeNode refChild, TreeNode newChild) {
        int position = children.indexOf(refChild);
        if (position == -1)  {
            throw new NoSuchChildException("Reference node is not a child of this node.");   
        }
        insertChild(position, newChild);
    }

    /**
     * Inserts a new child immediately after an existing child.
     * If <code>refChild</code> is not a child of this node, then
     * a <code>NoSuchChildException</code> is thrown. 
     * 
     * @param refChild the node after which the new node will be inserted
     * @param newChild the node inserted into the tree
     * 
     * @throws NoSuchChildException if <code>refChild</code> is not a child of this node
     * @throws NullPointerException if <code>newChild</code> is null
     * @throws MultipleParentException if <code>newChild</code> already has a parent
     * @throws IllegalAddException if this node cannot have children of this type
     */
    public final void insertAfter(TreeNode refChild, TreeNode newChild) {
        int position = children.indexOf(refChild);
        if (position == -1)  {
            throw new NoSuchChildException(
              "Reference node is not a child of this node.");   
        }
        insertChild(position+1, newChild);
    }

    /**
     * Returns the first child of this node, or null if
     * this node does not have any children.
     * 
     * @return TreeNode - thr first child of this node
     */
    public TreeNode getFirstChild() {
        try {
            return getChild(0); 
        }
        catch (IndexOutOfBoundsException e) {
            return null;   
        }
    }

    /**
     * Returns the child of this node at the specified position.
     * Indexes begin at 0 and count up to one less than the number
     * of children in this node.
     * 
     * @param position index of the node to return
     * 
     * @return the node at the requestd position
     * 
     * @throws IndexOutOfBoundsException if the index is negative or greater than
     *                                     the number of children of the node - 1.
     */
    public TreeNode getChild(int position) {
        return (TreeNode) (children.get(position)); 
    }

    /**
     * Returns the next node following this node in the tree
     * with the same parent as this node.
     * It returns null, if this  node is the last child 
     * of its parent, or if this node does not have a parent
     * (in which case it cannot have any siblings.)
     * 
     * @return the node which precedes this one
     */
    public TreeNode getNextSibling() {
        TreeNode parent = getParent();
        if (parent == null) return null;
        
        List siblings = parent.children;
        int position = siblings.indexOf(this);
        if (position == siblings.size()-1) return null;
        else return (TreeNode) siblings.get(position+1);
    }

    /**
     * Returns the node immediately preceding this node in the tree
     * with the same parent as this node.
     * It returns null, if this  node is the first child 
     * of its parent, or if this node does not have a parent
     * (in which case it cannot have any siblings.)
     * 
     * @return the node which precedes this one
     */
    public TreeNode getPreviousSibling() {
        TreeNode parent = getParent();
        if (parent == null) return null;
        
        List siblings = parent.children;
        int position = siblings.indexOf(this);
        if (position == 0) return null;
        else return (TreeNode) siblings.get(position-1);
    }

    /**
     * Removes the child of this node at the specified position.
     * Indexes begin at 0 and count up to one less than the number
     * of children in this node.
     * 
     * @param position index of the node to remove
     * 
     * @throws IndexOutOfBoundsException if the index is negative or greater than
     *                                     the number of children of this node - 1.
     */
    public void removeChild(int position) {
        TreeNode removed = (TreeNode) children.remove(position);
        removed.setParent(null);    
    }

    /**
     * Removes the specified child of this node.
     * 
     * @param child child node to remove
     * 
     * @throws NoSuchChildException if <code>child</code> is not in fact a child of this node.
     */
    public void removeChild(TreeNode child) {
        if (children.contains(child)) {
            children.remove(child);
            child.setParent(null);
        }
        else {
            throw new NoSuchChildException("Child does not belong to this node");
        }
    }

    /**
     * Detaches all children from this node.
     */
    public void removeChildren() {
        Iterator iterator = children.iterator();
        while (iterator.hasNext()) {
            TreeNode node = (TreeNode) iterator.next();
            node.setParent(null);   
        }
        children.clear();   
    }


    /**
     * Replaces an existing child with a new child node.
     * If <code>oldChild</code> is not a child of this node, then
     * a <code>NoSuchChildException</code> is thrown. 
     * 
     * @param oldChild the node removed from the tree
     * @param newChild the node inserted into the tree
     * 
     * @throws NoSuchChildException if <code>oldChild</code> is not a child of this node
     * @throws NullPointerException if <code>newChild</code> is null
     * @throws IllegalAddException if this node cannot have children of the type of <code>newChild</code>
     */
    public final void replaceChild(TreeNode oldChild, TreeNode newChild) {
        int position = children.indexOf(oldChild);
        if (position == -1)  {
            throw new NoSuchChildException("Reference node is not a child of this node.");   
        }
        removeChild(position);
        insertChild(position, newChild);
    }
    // do I need a checkInsert method for subclasses????
    // and possibly a check remove????


    /**
     * <p>
     * Returns a deep copy of this tree node with no parent,
     * that can be added to this document or a different one.
     * </p>
     * 
     * <p>
     * Per Bloch, the Cloneable interface is just a mess and should 
     * be avoided. However, I do not follow his suggestion of a copy
     * constructor because it is useful to be able to
     * copy a node without knowing its more specific type.
     * </p> 
     * 
     */
    public abstract TreeNode copy();        


}
