/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom;

import java.util.Stack;

import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.DTDHandler;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.ext.DeclHandler;

/**
 * @author Elliotte Rusty Harold
 * @version 1.0d2
 *
 */
class XOMHandler 
 implements ContentHandler, LexicalHandler, DeclHandler, DTDHandler {

    private Document    document;
    private Stack       parents;
    private boolean     inProlog;
    private int         position; // number of items currently in prolog
    private Locator     locator; // not used yet
    private DocType     doctype;
    private NodeFactory factory;
    
    /**
     * Method XOMHandler.
     * 
     * @param factory
     */
    XOMHandler(NodeFactory factory) {
        if (factory == null) throw new NullPointerException("Factory cannot be null");
        this.factory = factory; 
    }   
    
    /**
     * @see org.xml.sax.ContentHandler#setDocumentLocator(Locator)
     */
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    /**
     * Method getDocument.
     * 
     * @return Document
     */
    Document getDocument() {
        return document;
    }

    public void startDocument() {
        parents = new Stack();
        Element fauxRoot = new Element("fauxRoot");
        document = factory.makeDocument(fauxRoot);
        parents.push(document); 
        inProlog = true;
        position = 0;
        buffer = new StringBuffer();
        doctype = null;
    }
  
    public void endDocument() {
        parents.pop();
    }
  
    public void startElement(String namespaceURI, String localName, 
     String qualifiedName, org.xml.sax.Attributes attributes) {
        
        flushText();
        int colon = qualifiedName.indexOf(':');
        Element element = factory.makeElement(qualifiedName, namespaceURI);
        TreeNode parent = (TreeNode) parents.peek();
        if (parent instanceof Document) {
            Document doc = (Document) parent;
            doc.setRootElement(element);
            inProlog = false;
        }
        else {
            parent.appendChild(element);
        }
        parents.push(element);
        
        // Attach the attributes
        for (int i = 0; i < attributes.getLength(); i++) {
            String qName = attributes.getQName(i);
            // handle namespaces later
            if (qName.equals("xmlns") || qName.startsWith("xmlns:")) continue;
            String namespace = attributes.getURI(i);
            String value = attributes.getValue(i);
            Attribute attribute = factory.makeAttribute(
              qName, namespace, value, getType(attributes.getType(i))
            );
            element.add(attribute);    
        }

        // Attach any additional namespaces
        for (int i = 0; i < attributes.getLength(); i++) {
            String qName = attributes.getQName(i);
            if (qName.startsWith("xmlns:")) {
                
                String namespaceName = attributes.getValue(i);
                String namespacePrefix = qName.substring(6);
                String currentValue = element.getNamespaceURI(namespacePrefix); 
                if (!namespaceName.equals(currentValue)) {
                    element.addNamespaceDeclaration(namespacePrefix, namespaceName);
                }
                 
            }   
        }

    }
    
    private Attribute.Type getType(String saxType) {
    
        if (saxType.equals("CDATA")) return Attribute.Type.CDATA;
        if (saxType.equals("ID")) return Attribute.Type.ID;
        if (saxType.equals("IDREF")) return Attribute.Type.IDREF;
        if (saxType.equals("IDREFS")) return Attribute.Type.IDREFS;
        if (saxType.equals("NMTOKEN")) return Attribute.Type.NMTOKEN;
        if (saxType.equals("NMTOKENS")) return Attribute.Type.NMTOKENS;
        if (saxType.equals("ENTITY")) return Attribute.Type.ENTITY;
        if (saxType.equals("ENTITIES")) return Attribute.Type.ENTITIES;
        if (saxType.equals("NOTATION")) return Attribute.Type.NOTATION;
        // non-standard but some parsers use this
        if (saxType.equals("ENUMERATION")) return Attribute.Type.ENUMERATION;
        if (saxType.startsWith("(")) return Attribute.Type.ENUMERATION;
    
        return Attribute.Type.UNDECLARED;
        
    }
  
    public void endElement(String namespaceURI, String localName, String qualifiedName) {
        flushText();
        parents.pop(); 
    }
  
    private StringBuffer buffer;
  
    public void characters(char[] text, int start, int length) {
        buffer.append(text, start, length); 
    }
 
    // acumulate all text that's in the buffer into a text node
    private void flushText() {
        if (buffer.length() > 0) {
            Text data = factory.makeText(buffer.toString());
            TreeNode parent = (TreeNode) parents.peek();
            parent.appendChild(data);
            buffer.setLength(0);
        }       
    }
  
    public void ignorableWhitespace(char[] text, int start, int length) {
        characters(text, start, length);
    }
  
    public void processingInstruction(String target, String data) {
        if (!inDTD) {
            flushText();
            ProcessingInstruction instruction 
             = factory.makeProcessingInstruction(target, data);
            TreeNode parent = (TreeNode) parents.peek();
            if (inProlog) {
                parent.insertChild(position, instruction);
                position++;
            }
            else {
                parent.appendChild(instruction);
            }
        }
        else if (!inExternalSubset) {
            doctype.addProcessingInstruction("<? " + target + " " + data + "?>");            
        }

    }


    // I handle this with attribute values; not prefix mappings
    public void startPrefixMapping(String prefix, String uri) {}
    public void endPrefixMapping(String prefix) {}

    public void skippedEntity(String name) {
        flushText();
        throw new XMLException("Could not resolve entity " + name);                         
    }
    
    // LexicalHandler events
    
    private boolean inDTD = false;

    public void startDTD(String rootName, String publicID, 
      String systemID) {
        inDTD = true;
        DocType doctype = new DocType(rootName, publicID, systemID);    
        TreeNode parent = (TreeNode) parents.peek();
        parent.insertChild(position, doctype);
        position++;
        this.doctype = doctype;
     }
     
    public void endDTD() {
        inDTD = false;
    }

    private boolean inExternalSubset = false;

    public void startEntity(String name) {
      if (name.equals("[dtd]")) inExternalSubset = true;
    }
    
    public void endEntity(String name) {
      if (name.equals("[dtd]")) inExternalSubset = false;    
    }
    
    public void startCDATA() {}
    public void endCDATA() {}

    public void comment(char[] text, int start, int length) {
        if (!inDTD && !inExternalSubset) {
            flushText();
            Comment comment = factory.makeComment(new String(text, start, length));
            TreeNode parent = (TreeNode) parents.peek();
            if (inProlog) {
                parent.insertChild(position, comment);
                position++;
            }
            else {
                parent.appendChild(comment);
            }
        }
        else if (inDTD && !inExternalSubset) {
            doctype.addComment("<!-- " + (new String(text, start, length)) + "-->");            
        }
    } 
    
    
    public void elementDecl(String name, String model) {
        if (!inExternalSubset ) {
            doctype.addDeclaration("<!ELEMENT " + name + " " + model + " >");
        }
    }
  
    public void attributeDecl(String elementName, 
      String attributeName, String type, String mode, 
      String defaultValue)  {
    
        if (!inExternalSubset ) {
            StringBuffer result = new StringBuffer();
        
            result.append("<!ATTLIST ");
            result.append(elementName);
            result.append(" ");
            result.append(attributeName);
            result.append(" ");
            result.append(type);
            result.append(" ");
            if (mode != null) {
                result.append(mode + " ");
            }
            if (defaultValue != null) {
                result.append('"' + defaultValue + "\" ");
            }
            result.append(">");   
            doctype.addDeclaration(result.toString());
        }
    }
  
    public void internalEntityDecl(String name, 
       String value) {   
        if (!inExternalSubset ) {
            if (!name.startsWith("%")) { // ignore parameter entities
                doctype.addDeclaration("<!ENTITY " + name + " \"" + value + "\">");        
            }
        }
    }
  
    public void externalEntityDecl(String name, 
       String publicID, String systemID) {
     
        if (!inExternalSubset ) {
            if (!name.startsWith("%")) { // ignore parameter entities
                if (publicID != null) { 
                    doctype.addDeclaration("<!ENTITY " + name + " PUBLIC \"" 
                      + publicID + "\" \"" + systemID + "\">");        
                }
                else {
                    doctype.addDeclaration("<!ENTITY " + name + " SYSTEM \"" 
                      + systemID + "\">");        
                }
            }
        }
    }
    
    public void notationDecl(String name, String publicID, 
      String systemID) {
        
        if (!inExternalSubset ) {
            doctype.addDeclaration("<!NOTATION " + name + " PUBLIC \"" + publicID 
             + "\" \"" + systemID + "\">");
        }        
        
    }
   
    public void unparsedEntityDecl(String name, String publicID, 
     String systemID, String notationName) {
        if (!inExternalSubset ) {
            if (publicID != null) { 
                doctype.addDeclaration("<!ENTITY " + name + " PUBLIC \"" 
                  + publicID + "\" \"" + systemID + "\" NDATA " + notationName + ">");        
            }
            else {
                doctype.addDeclaration("<!ENTITY " + name + " SYSTEM \"" 
                  + systemID + "\" NDATA " + notationName + ">");        
            }
        }
        
    }
 
}