/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom.tests;

import junit.framework.Assert;
import junit.framework.TestCase;

import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.IllegalDataException;
import nu.xom.NamespaceException;

/**
 * @author Elliotte Rusty Harold
 *
 */
public class AttributeTest extends TestCase {

    public AttributeTest() {
        super("Attribute tests");
    }

    public AttributeTest(String name) {
        super(name);
    }

    private Attribute a1;
    private Attribute a2;


    protected void setUp() {
        a1 = new Attribute("test", "value");
        a2 = new Attribute("test", "  value  ");
    }

    public void testConstructor() {
        Assert.assertEquals(a1.getLocalName(), "test");
        Assert.assertEquals(a1.getQualifiedName(), "test");
        Assert.assertEquals(a1.getNamespacePrefix(), "");
        Assert.assertEquals(a1.getNamespaceURI(), "");
        Assert.assertEquals(a1.getValue(), "value");
        Assert.assertEquals(a2.getValue(), "  value  ");
    }

    public void testConstructor2() {
        
        Attribute a1 = new Attribute("name", "value", Attribute.Type.CDATA);
        Assert.assertEquals(a1.getLocalName(), "name");
        Assert.assertEquals(a1.getQualifiedName(), "name");
        Assert.assertEquals(a1.getNamespacePrefix(), "");
        Assert.assertEquals(a1.getNamespaceURI(), "");
        Assert.assertEquals(a1.getValue(), "value");
        Assert.assertEquals(a1.getType(), Attribute.Type.CDATA);
    }

    public void testGetExternalForm() {
    	
        Attribute a1 = new Attribute("test", "value contains a \"");
        Assert.assertEquals("test=\"value contains a &quot;\"", a1.getStringForm());

        Attribute a2 = new Attribute("test", "value contains a '");
        Assert.assertEquals("test=\"value contains a &apos;\"", a2.getStringForm());


     }

    // No xmlns attributes or xmlns:prefix attributes are allowed
    public void testXmlns() {
    	
    	try {
           	new Attribute("xmlns", "http://www.w3.org/TR");
        	fail("Created attribute with name xmlns");
    	}
    	catch (NamespaceException ex) {
    		// success	
    	}
 
    	try {
           	new Attribute("xmlns:prefix", "http://www.w3.org/TR");
        	fail("Created attribute with name xmlns:prefix");
    	}
    	catch (NamespaceException ex) {
    		// success	
    	}
 
		// Now try with namespace URI from errata
		try {
 			new Attribute("xmlns", "http://www.w3.org/2000/xmlns/", "http://www.w3.org/");
			fail("created xmlns attribute");
 		}
 		catch (NamespaceException e) {
 			// success	
 		}
		
		// Now try with namespace URI from errata
		try {
 			new Attribute("xmlns:pre", "http://www.w3.org/2000/xmlns/", "http://www.w3.org/");
			fail("created xmlns:pre attribute");
 		}
 		catch (NamespaceException e) {
 			// success	
 		}

     }

    public void testXmlPrefix() {
    	
    	try {
           	new Attribute("xml:base", "http://www.w3.org/TR");
        	fail("Created attribute with name xml:base");
    	}
    	catch (NamespaceException ex) {
    		// success	
    	}
 
    	try {
           	new Attribute("xml:space", "preserve");
        	fail("Created attribute with name xml:space");
    	}
    	catch (NamespaceException ex) {
    		// success	
    	}
 
    	try {
           	new Attribute("xml:lang", "fr-FR");
        	fail("Created attribute with name xml:lang");
    	}
    	catch (NamespaceException ex) {
    		// success	
    	}
        
        String xmlNamespace = "http://www.w3.org/XML/1998/namespace";    	
        Attribute a1 = new Attribute("xml:base", xmlNamespace, "http://www.w3.org/");
        Assert.assertEquals(a1.getLocalName(), "base");
        Assert.assertEquals(a1.getQualifiedName(), "xml:base");
        Assert.assertEquals(a1.getNamespaceURI(), xmlNamespace);

        Attribute a2 = new Attribute("xml:space", xmlNamespace, "preserve");
        Assert.assertEquals(a2.getLocalName(), "space");
        Assert.assertEquals(a2.getQualifiedName(), "xml:space");
        Assert.assertEquals(a2.getNamespaceURI(), xmlNamespace);

        Attribute a3 = new Attribute("xml:lang", xmlNamespace, "en-UK");
        Assert.assertEquals(a3.getLocalName(), "lang");
        Assert.assertEquals(a3.getQualifiedName(), "xml:lang");
        Assert.assertEquals(a3.getNamespaceURI(), xmlNamespace);

 		try {
 			new Attribute("xml:base", "http://www.nothteXMLNamespace", "http://www.w3.org/");
			fail("remapped xml prefix");
 		}
 		catch (NamespaceException e) {
 			// success	
 		}

     }




    public void testSetValue() {
    	
    	String[] legal = {
    	  "Hello",
    	  "hello there",
    	  "  spaces on both ends  ",
    	  " quotes \" \" quotes",
    	  " single \'\' quotes",
    	  " both double and single \"\'\"\' quotes",  
    	  " angle brackets <  > <<<",  
    	  " carriage returns \r\r\r",  
    	  " ampersands & &&& &name; "  
    	};

        // Things that shouldn't cause an exception
        for (int i = 0; i < legal.length; i++) {
        	a1.setValue(legal[i]);	
        	Assert.assertEquals(a1.getValue(), legal[i]);
        }
        
        try {
          a1.setValue("test \u0000 test ");
          fail("Should raise an IllegalDataException");
        }
        catch (IllegalDataException success) {}


     }

    public void testNames() {
    	
        String prefix = "testPrefix";
        String name = "testName";
        String URI = "http://www.elharo.com/";
        String value = "  here's some data";
        
    	
        Attribute a1 = new Attribute(prefix + ":" + name, URI, value);
        Assert.assertEquals(a1.getLocalName(), name);
        Assert.assertEquals(a1.getQualifiedName(), prefix + ":" + name);
        Assert.assertEquals(a1.getNamespaceURI(), URI);
    }


    public void testEquals() {
        Attribute c1 = new Attribute("test", "limit");
        Attribute c2 = new Attribute("test", "limit");
        Attribute c3 = new Attribute("skjlchsakdjh", "sajdh sajh ");

        Assert.assertEquals(c1, c1);
        Assert.assertEquals(c1.hashCode(), c1.hashCode());
        Assert.assertTrue(!c1.equals(c2));
        Assert.assertTrue(!c1.equals(c3));
    }

    public void testClone() {
        Attribute c1 = new Attribute("test", "data");
        Attribute c2 = new Attribute(c1);

        Assert.assertEquals(c1.getValue(), c2.getValue());
        Assert.assertEquals(c1.getLocalName(), c2.getLocalName());
        Assert.assertEquals(c1.getQualifiedName(), c2.getQualifiedName());
        Assert.assertEquals(c1.getValue(), c2.getValue());
        Assert.assertTrue(!c1.equals(c2));
        Assert.assertNull(c2.getParent());

    }

    // Check passing in a string with broken surrogate pairs
    // and with correct surrogate pairs
    public void testSurrogates() {

        String goodString = "test: \uD8F5\uDF80  ";
        Attribute c = new Attribute("surrogate", goodString);
        Assert.assertEquals(goodString, c.getValue());

        // Is it possible and legal for a surrogate pair to
        // decode to something in the BMP????

        // Two high-halves
        try {
          Attribute c1 = new Attribute("surrogate", "test: \uD8F5\uDBF0  ");
          fail("Should raise an IllegalDataException");
        }
        catch (IllegalDataException success) {}


        // Two high-halves
        try {
          Attribute c1 = new Attribute("surrogate", "test: \uD8F5\uD8F5  ");
          fail("Should raise an IllegalDataException");
        }
        catch (IllegalDataException success) {}

        // One high-half
        try {
           Attribute c1 = new Attribute("surrogate", "test: \uD8F5  ");
           fail("Should raise an IllegalDataException");
         }
         catch (IllegalDataException success) {}

        // One low half
         try {
            Attribute c1 = new Attribute("surrogate", "test: \uDF80  ");
            fail("Should raise an IllegalDataException");
          }
          catch (IllegalDataException success) {}

        // Low half before high half
         try {
            Attribute c1 = new Attribute("surrogate", "test: \uDCF5\uD8F5  ");
            fail("Should raise an IllegalDataException");
          }
          catch (IllegalDataException success) {}


    }


    // :learn to use todo????

    public void testLeafNode() {

        Attribute a1 = new Attribute("test", "data");

        Assert.assertNull(a1.getParent());

        Element element = new Element("test");
        element.add(a1);
        Assert.assertEquals(a1.getParent(), element);
        Assert.assertEquals(a1, element.getAttribute("test"));

        element.removeAttribute(a1);
        Assert.assertNull(element.getAttribute("test"));

    }

}