/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom.tests;

import junit.framework.TestCase;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;

import nu.xom.DocType;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Attribute;
import nu.xom.DOMConverter;
import nu.xom.ProcessingInstruction;
import nu.xom.Comment;
import org.w3c.dom.DOMImplementation;
import javax.xml.parsers.*;
import org.apache.xerces.parsers.DOMParser;
import org.xml.sax.InputSource;


/**
 * @author Elliotte Rusty Harold
 *
 */
public class DOMConverterTester extends TestCase {

    public DOMConverterTester(String name) {
        super(name);   
    }

    private String source = "<!DOCTYPE test>\r\n"
     + "<?xml-stylesheet href=\"file.css\" type=\"text/css\"?>" 
     + "<!-- test -->"
     + "<test xmlns:xlink='http://www.w3.org/TR/1999/xlink'>Hello dear"
     + "\r\n<em id=\"p1\" xmlns:none=\"http://www.example.com\">very important</em>"
     + "<span xlink:type='simple'>here&apos;s the link</span>\r\n"
     + "<svg:svg xmlns:svg='http://www.w3.org/TR/2000/svg'><svg:text>text in a namespace</svg:text></svg:svg>\r\n"
     + "<svg xmlns='http://www.w3.org/TR/2000/svg'><text>text in a namespace</text></svg>"
     + "</test>\r\n"
     + "<!--epilog-->";
     
    private Document xomDocument;
    private org.w3c.dom.Document domDocument;
    private DOMImplementation impl;


    protected void setUp() throws ParserConfigurationException {
               
        DocType type = new DocType("test");
        Element root = new Element("test");          
        xomDocument = new Document(root);
        xomDocument.insertBefore(root, type);
        xomDocument.insertBefore(root, new ProcessingInstruction("xml-stylesheet", 
         "href=\"file.css\" type=\"text/css\""));
        xomDocument.insertBefore(root, new Comment(" test "));
        xomDocument.appendChild(new Comment("epilog"));
        root.addNamespaceDeclaration("xlink", "http://www.w3.org/TR/1999/xlink");
        root.appendChild("Hello dear\r\n");
        Element em = new Element("em");
        root.appendChild(em);
        em.add(new Attribute("id", "p1"));
        em.addNamespaceDeclaration("none", "http://www.example.com");
        em.appendChild("very important");
        Element span = new Element("span");
        root.appendChild(span);
        span.add(new Attribute("xlink:type", "http://www.w3.org/TR/1999/xlink", "simple"));
        span.appendChild("here's the link");
        root.appendChild("\r\n");

        Element empty = new Element("empty");
        root.appendChild(empty);
        empty.add(new Attribute("xom:temp", "http://xom.nu/", "Just to see if this can handle namespaced attributes"));
        
        root.appendChild("\r\n");
        Element svg = new Element("svg:svg", "http://www.w3.org/TR/2000/svg");
        root.appendChild(svg);
        Element text = new Element("svg:text", "http://www.w3.org/TR/2000/svg");
        svg.appendChild(text);
        text.appendChild("text in a namespace");
        root.appendChild("\r\n");

        svg = new Element("svg", "http://www.w3.org/TR/2000/svg");
        root.appendChild(svg);
        text = new Element("text", "http://www.w3.org/TR/2000/svg");
        svg.appendChild(text);
        text.appendChild("text in a namespace");
        
        Reader reader = new StringReader(source);
        DocumentBuilderFactory factory 
          = DocumentBuilderFactory.newInstance();
        impl = factory.newDocumentBuilder().getDOMImplementation();        
        InputSource inso = new InputSource(new StringReader(source));
        DOMParser parser = new DOMParser();
        try {
            parser.parse(inso);
            domDocument = parser.getDocument();
        }
        catch (Exception ex) {
            // shouldn't happen from known good doc 
            throw new RuntimeException("Ooops!");  
        }
    }

    public void testToDOM() throws IOException {
        
        org.w3c.dom.Document domDoc = DOMConverter.translate(xomDocument, impl);
        org.w3c.dom.DocumentType doctype = domDoc.getDoctype();
        
        assertEquals(domDoc.getDocumentElement().getNodeName(), "test");
        assertTrue(doctype != null);
        assertEquals(domDoc.getFirstChild().getNodeType(), org.w3c.dom.Node.DOCUMENT_TYPE_NODE);
        assertEquals(domDoc.getLastChild().getNodeType(), org.w3c.dom.Node.COMMENT_NODE);
        
      /*  org.apache.xml.serialize.OutputFormat format = new org.apache.xml.serialize.OutputFormat();
        org.apache.xml.serialize.DOMSerializer serializer 
          = new org.apache.xml.serialize.XMLSerializer(System.out, format);
        serializer.serialize(domDoc);
        
        System.out.flush(); */
    }

    public void testToXOM() {
        
        Document xomDoc = DOMConverter.translate(domDocument);
        DocType doctype = xomDoc.getDocType();
        Element root = xomDoc.getRootElement();

        assertEquals(root.getQualifiedName(), "test");
        assertEquals(root.getLocalName(), "test");
        assertEquals(root.getNamespaceURI(), "");
        
        assertTrue(doctype != null);
        assertTrue(xomDoc.getFirstChild() instanceof DocType);
        assertTrue(root.getNextSibling() instanceof nu.xom.Comment);
        assertTrue(root.getPreviousSibling() instanceof nu.xom.Comment);
        assertEquals(root.getPreviousSibling().getValue(), " test ");
        assertEquals(root.getNextSibling().getValue(), "epilog");
        assertTrue(doctype.getNextSibling() instanceof nu.xom.ProcessingInstruction);
        assertEquals(doctype.getRootElementName(), "test");
        assertNull(doctype.getPublicID());
        assertNull(doctype.getSystemID());
        
        // System.out.println(xomDoc.getStringForm());
       
    }


}
