/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom.tests;

import junit.framework.TestCase;
import junit.framework.Assert;

import nu.xom.Document;
import nu.xom.Element;
import nu.xom.ProcessingInstruction;
import nu.xom.Comment;
import nu.xom.Text;
import nu.xom.IllegalAddException;
import nu.xom.WellformednessException;
import nu.xom.NoSuchChildException;


/**
 * @author Elliotte Rusty Harold
 *
 */
public class DocumentTest extends TestCase {

    /**
     * Constructor for DocumentTest.
     * @param name
     */
    public DocumentTest(String name) {
        super(name);
    }
    
    Element root;
    Document doc;
    
    
    protected void setUp() {
        root = new Element("root");
        doc = new Document(root);

    }

    public void testSetRoot() {
        Element newRoot = new Element("newroot");
        doc.setRootElement(newRoot);
        
        assertEquals(doc.getRootElement(), newRoot);
        assertNull(root.getParent());
        
    }

    public void testGetDocument() {
        assertEquals(doc, doc.getDocument());
    }

    public void testConstructor() {
    
        Element root = new Element("root");
        Document doc = new Document(root);

        Assert.assertEquals(root, doc.getRootElement());
        Assert.assertEquals(1, doc.getChildCount());
        
        Element newRoot = new Element("newRoot");
        doc.setRootElement(newRoot);
        Assert.assertEquals(newRoot, doc.getRootElement());
        Assert.assertEquals(1, doc.getChildCount());
        
        doc.appendChild(new Comment("test"));
        Assert.assertEquals(2, doc.getChildCount());

        doc.insertChild(0, new Comment("prolog comment"));
        Assert.assertEquals(3, doc.getChildCount());
        Assert.assertTrue(doc.getFirstChild() instanceof Comment);
        Assert.assertTrue(doc.getChild(1) instanceof Element);
        Assert.assertTrue(doc.getChild(2) instanceof Comment);

        doc.insertChild(1, new ProcessingInstruction("target", "data"));
        Assert.assertTrue(doc.getFirstChild() instanceof Comment);
        Assert.assertTrue(doc.getChild(1) instanceof ProcessingInstruction);
        Assert.assertTrue(doc.getChild(2) instanceof Element);
        Assert.assertTrue(doc.getChild(3) instanceof Comment);

        doc.insertChild(3, new ProcessingInstruction("epilog", "data"));
        Assert.assertTrue(doc.getFirstChild() instanceof Comment);
        Assert.assertTrue(doc.getChild(1) instanceof ProcessingInstruction);
        Assert.assertTrue(doc.getChild(2) instanceof Element);
        Assert.assertTrue(doc.getChild(3) instanceof ProcessingInstruction);
        Assert.assertTrue(doc.getChild(4) instanceof Comment);

        
        try {
            Element nullRoot = null;
            new Document(nullRoot);
            fail("allowed null root!");
        }
        catch (NullPointerException ex) {
            // success  
        }
        
        try {
            Document nullDoc = null;
            new Document(nullDoc);
            fail("allowed null root!");
        }
        catch (NullPointerException ex) {
            // success  
        }
        
    }
    
    public void testAppend() {
    
        Element root = new Element("root");
        Document doc = new Document(root);

        try {
            doc.appendChild("text");
            fail("appended string");
        }   
        catch (IllegalAddException e) {
            // success  
        }
        
        try {
            doc.appendChild("    ");
            fail("appended white space");
        }   
        catch (IllegalAddException e) {
            // success  
        }
        
        try {
            doc.appendChild(new Text("test"));
            fail("appended text");
        }   
        catch (IllegalAddException e) {
            // success  
        }
        
        try {
            doc.appendChild(new Comment("test"));
            doc.appendChild(new Element("test"));
            fail("appended element");
        }   
        catch (IllegalAddException e) {
            // success  
        }
        
        try {
            doc.insertChild(0, new Element("test"));
            fail("inserted element");
        }   
        catch (IllegalAddException e) {
            // success  
        }
        
    }

    public void testRemoval() {
    
        Element root = new Element("root");
        Document doc = new Document(root);

        try {
            root.detach();
            fail("detached root element");
        }   
        catch (WellformednessException e) {
            // success  
        }
        
        try {
            doc.removeChild(root);
            fail("removed root element");
        }   
        catch (WellformednessException e) {
            // success  
        }
        
        try {
            doc.removeChild(0);
            fail("removed root element");
        }   
        catch (WellformednessException e) {
            // success  
        }
        
        doc.appendChild(new Comment("test"));
        doc.removeChild(1);
        Assert.assertEquals(1, doc.getChildCount());
        
        Comment test = new Comment("test");
        doc.appendChild(test);
        doc.removeChild(test);
        Assert.assertEquals(1, doc.getChildCount());
        
        try {
            Comment something = new Comment("sd");
            doc.removeChild(something);
            fail("Removed nonchild");
        }
        catch (NoSuchChildException ex) {
            // success   
        }

        try {
            doc.removeChild(20);
            fail("removed overly sized element");
        }   
        catch (IndexOutOfBoundsException e) {
            // success  
        }


        
    }



}
