/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom.tests;

/**
 * @author Elliotte Rusty Harold
 *
 */
/*
 */

import junit.framework.TestCase;
import junit.framework.Assert;
import nu.xom.Element;
import nu.xom.Text;
import nu.xom.Document;
import nu.xom.Comment;
import nu.xom.ProcessingInstruction;
import nu.xom.Attribute;
import nu.xom.Attributes;
import nu.xom.IllegalDataException;
import nu.xom.IllegalAddException;
import nu.xom.IllegalNameException;
import nu.xom.MultipleParentException;
import nu.xom.NamespaceException;


public class ElementTest extends TestCase {

    public ElementTest() {
        super("Element tests");
    }

    public ElementTest(String name) {
        super(name);
    }

    public void testConstructor1() {
        String name = "sakjdhjhd";
        Element e = new Element(name);
        
        Assert.assertEquals(name, e.getLocalName());
        Assert.assertEquals(name, e.getQualifiedName());
        Assert.assertEquals("", e.getNamespacePrefix());
        Assert.assertEquals("", e.getNamespaceURI());
    }

    public void testConstructor2() {
        String name = "sakjdhjhd";
        String uri = "http://www.something.com/";
        Element e = new Element(name, uri);
        
        Assert.assertEquals(name, e.getLocalName());
        Assert.assertEquals(name, e.getQualifiedName());
        Assert.assertEquals("", e.getNamespacePrefix());
        Assert.assertEquals(uri, e.getNamespaceURI());
    }

   public void testConstructor3() {
        String name = "red:sakjdhjhd";
        String uri = "http://www.something.com/";
        Element e = new Element(name, uri);
        
        Assert.assertEquals("sakjdhjhd", e.getLocalName());
        Assert.assertEquals(name, e.getQualifiedName());
        Assert.assertEquals("red", e.getNamespacePrefix());
        Assert.assertEquals(uri, e.getNamespaceURI());
    }


    public void testXMLNamespace() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        Element e = new Element(name, uri);
        
        String xmlNamespace = "http://www.w3.org/XML/1998/namespace";
        
        Assert.assertEquals(e.getNamespaceURI("xml"), xmlNamespace);
        
        try {
            e.addNamespaceDeclaration("xml", "http://www.yahoo.com/");
            fail("remapped xml prefix!");
        }
        catch (NamespaceException ex) {
            // success
        }
        Assert.assertEquals(e.getNamespaceURI("xml"), xmlNamespace);
        
        // This should succeed
        e.addNamespaceDeclaration("xml", xmlNamespace);
        Assert.assertEquals(e.getNamespaceURI("xml"), xmlNamespace);

        
    }

    public void testNamespaceMappings() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        Element e = new Element(name, uri);
        e.addNamespaceDeclaration("blue", "http://www.blue.com/");
        e.addNamespaceDeclaration("green", "http://www.green.com/");
        Attribute a1 = new Attribute("test", "test");
        Attribute a2 = new Attribute("pre1:green", "http://www.green.com/", "data");
        Attribute a3 = new Attribute("yellow:sfsdadf", "http://www.yellow.com/", "data");
        e.add(a1);
        e.add(a2);
        e.add(a3);
        
        Assert.assertEquals("http://www.red.com/", e.getNamespaceURI("red"));
        Assert.assertEquals("http://www.green.com/", e.getNamespaceURI("green"));
        Assert.assertEquals("http://www.blue.com/", e.getNamespaceURI("blue"));
        Assert.assertEquals("http://www.green.com/", e.getNamespaceURI("pre1"));
        Assert.assertEquals("http://www.yellow.com/", e.getNamespaceURI("yellow"));
        
        
        Element e2 = new Element("mauve:child", "http://www.mauve.com");
        e.appendChild(e2);
        Assert.assertEquals("http://www.red.com/", e2.getNamespaceURI("red"));
        Assert.assertEquals("http://www.blue.com/", e2.getNamespaceURI("blue"));
        Assert.assertEquals("http://www.green.com/", e2.getNamespaceURI("green"));
        Assert.assertEquals("http://www.green.com/", e2.getNamespaceURI("pre1"));
        Assert.assertEquals("http://www.yellow.com/", e2.getNamespaceURI("yellow"));
        Assert.assertNull(e2.getNamespaceURI("head"));
        
        
        try {
            
            e.addNamespaceDeclaration("pre1", "http://www.blue2.com");

            fail("Added conflicting namespace");    
        }
        catch (NamespaceException ex) {
          // success    
        }
        try {
            Attribute a4 = new Attribute("pre1:green", "http://www.sadas.com/", "data");
            e.add(a4);
            fail("Added conflicting namespace");    
        }
        catch (NamespaceException ex) {
          // success    
        }
        
        
        // Should be able to override namespaces:
        e.addNamespaceDeclaration("green", "http://www.green2.com/");
        Assert.assertEquals("http://www.green2.com/", e.getNamespaceURI("green"));
        
        
        e.removeNamespaceDeclaration("green");
        Assert.assertNull(e.getNamespaceURI("green"));
        
    }

    public void testAttributes() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        Element e = new Element(name, uri);

        Attribute a1 = new Attribute("name", "simple");
        Attribute a2 = new Attribute("pre1:green", "http://www.green.com/", "data");

        e.add(a1);
        e.add(a2);

        Assert.assertEquals(a2, e.getAttribute("green", "http://www.green.com/"));
        Assert.assertEquals(a1, e.getAttribute("name"));
        Assert.assertEquals(a1, e.getAttribute("name", ""));
        Assert.assertEquals(a1.getParent(), e);
        Assert.assertEquals("simple", e.getAttribute("name").getValue());
        
        a1.detach();
        Assert.assertNull(a1.getParent());
        Assert.assertNull(e.getAttribute("name"));
        
        
        e.removeAttribute(a2);
        Assert.assertNull(a2.getParent());
        Assert.assertNull( e.getAttribute("green", "http://www.green.com/"));
        
    }



    public void testGetValue() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        Assert.assertEquals(e.getValue(), "");
        
        e.appendChild(new Text("data"));
        Assert.assertEquals(e.getValue(), "data");
        e.appendChild(new Text(" moredata"));
        Assert.assertEquals(e.getValue(), "data moredata");
        e.appendChild(new Comment(" moredata"));
        Assert.assertEquals(e.getValue(), "data moredata");
        e.appendChild(new ProcessingInstruction("target", " moredata"));
        Assert.assertEquals(e.getValue(), "data moredata");
        
        Element e2 = new Element("child");
        e.appendChild(e2);
        Assert.assertEquals(e.getValue(), "data moredata");
        e2.appendChild(new Text("something"));
        Assert.assertEquals(e.getValue(), "data moredatasomething");
        
    }

    public void testSetLocalName() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        Assert.assertEquals(e.getLocalName(), "sakjdhjhd");
        
        e.setLocalName("dude");
        Assert.assertEquals(e.getLocalName(), "dude");
        e.setLocalName("digits__");
        Assert.assertEquals(e.getLocalName(), "digits__");
        e.setLocalName("digits1234");
        Assert.assertEquals(e.getLocalName(), "digits1234");
        e.setLocalName("digits-z");
        Assert.assertEquals(e.getLocalName(), "digits-z");
        
        
        try {
            e.setLocalName("spaces ");
            fail("Local name allowed to contain spaces");   
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        try {
            e.setLocalName("digits:test");
            fail("Local name allowed to contain colon");    
        }
        catch (NamespaceException ex) {
          // Success    
        }
        try {
            e.setLocalName("digits!test");
            fail("Local name allowed to contain bang"); 
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        try {
            e.setLocalName("digits\u0000test");
            fail("Local name allowed to contain null"); 
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        
    }

    public void testSetNamespacePrefix() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        Assert.assertEquals(e.getNamespacePrefix(), prefix);
        
        e.setNamespacePrefix("dude");
        Assert.assertEquals(e.getNamespacePrefix(), "dude");
        e.setNamespacePrefix("digits__");
        Assert.assertEquals(e.getNamespacePrefix(), "digits__");
        e.setNamespacePrefix("digits1234");
        Assert.assertEquals(e.getNamespacePrefix(), "digits1234");
        e.setNamespacePrefix("digits-z");
        Assert.assertEquals(e.getNamespacePrefix(), "digits-z");
        e.setNamespacePrefix("");
        Assert.assertEquals(e.getNamespacePrefix(), "");
        
        try {
            e.setNamespacePrefix("spaces ");
            fail("namespace prefix allowed to contain spaces"); 
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        try {
            e.setNamespacePrefix("digits:test");
            fail("namespace prefix allowed to contain colon");  
        }
        catch (NamespaceException ex) {
          // Success    
        }
        try {
            e.setNamespacePrefix("digits!test");
            fail("namespace prefix allowed to contain bang");   
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        try {
            e.setNamespacePrefix("digits\u0000test");
            fail("namespace prefix allowed to contain null");   
        }
        catch (IllegalNameException ex) {
          // Success    
        }
        
    }

    public void testSetNamespaceURI() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        Assert.assertEquals(e.getNamespaceURI(), uri);
        
        String[] legal = {"http://www.is.edu/sakdsk#sjadh",
        "http://www.is.edu/sakdsk?name=value&name=head",
        "uri:isbn:0832473864",
        "fred",
        "http://www.is.edu/%20sakdsk#sjadh",
        ""};
         
        String[] illegal = {
          "http://www.is.edu/%sakdsk#sjadh",
          "http://www.is.edu/k\u0245kakdsk#sjadh",
          "!@#$%^&*()"
        }; 
        
        for (int i = 0; i < legal.length; i++) {
            e.setNamespaceURI(legal[i]);          
            Assert.assertEquals(e.getNamespaceURI(), legal[i]);
        }
        
        for (int i = 0; i < illegal.length; i++) {
            try {
                e.setNamespaceURI(illegal[i]);
                fail("illegal namespace URI allowed");  
            }
            catch (NamespaceException ex) {
               // Success   
            }
        }
        
    }

    public void testInsertChild() {
        
        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        Attribute a1 = new Attribute("test", "test");       
        
        Element e2 = new Element("mauve:child", "http://www.mauve.com");
        e.insertChild(0, e2);
        Element e3 = new Element("mauve:child", "http://www.mauve.com");
        e.insertChild(0, e3);
        Element e4 = new Element("mauve:child", "http://www.mauve.com");
        e3.insertChild(0, e4);
        
        Assert.assertEquals(e3, e.getFirstChild());
        
        try {
            Element root = new Element("root");
            Document doc = new Document(root);
            e.insertChild(0, doc);
            fail("Added document as child");    
        }
        catch (IllegalAddException ex) {
          // success    
        }
        try {
            e.insertChild(0, e2);
            fail("Added child twice");  
        }
        catch (MultipleParentException ex) {
          // success    
        }
        try {
            e.insertChild(0, e4);
            fail("Added child twice");  
        }
        catch (MultipleParentException ex) {
          // success    
        }
        try {
            e.insertChild(20, new Comment("test"));
            fail("Added comment after end");    
        }
        catch (IndexOutOfBoundsException ex) {
          // success    
        }
        try {
            e.insertChild(-20, new Comment("test"));
            fail("Added comment before start"); 
        }
        catch (IndexOutOfBoundsException ex) {
          // success    
        }
        
        
    }
    

    public void testAddAttribute() {
        
        Element element = new Element("name");
        Attribute a1 = new Attribute("name", "value");
        Attribute a2 = new Attribute("xlink:type", "http://www.w3.org/TR/1999/xlink", "sikmple");
        
        element.add(a1);
        element.add(a2);
        
        Attributes atts = element.getAttributes();
        Assert.assertEquals(2, atts.size());
        
        
        Element element2 = new Element("name");
        try {
            element2.add(a1);
            fail("added attribute with existing parent");
        }
        catch (IllegalAddException e) {
         // success   
        }
        
        a1.detach();
        element2.add(a1);
        
        a2.detach();
        Element funky = new Element("xlink:funky", "http://www.funky.org");
        try {
            funky.add(a2); 
            fail("added conflicting namespace"); 
        }
        catch (Exception e) {
            // success   
        }
        
        a2.detach();
        Element notasfunky = new Element("prefix:funky", "http://www.w3.org/TR/1999/xlink");
        notasfunky.add(a2);

 
        Attribute a3 = new Attribute("xlink:type", "http://www.w3.org/TR/1999/xlink", "simple");
        Attribute a4 = new Attribute("xlink:href", "http://www.w3.org/1998/xlink", "simple");
        Element test = new Element("test");
        try {
            test.add(a3); 
            test.add(a4); 
            fail("added conflicting attributes"); 
        }
        catch (Exception e) {
            // success   
        }

        Attribute a5 = new Attribute("xlink:type", "http://www.w3.org/TR/1999/xlink", "sikmple");
        Attribute a6 = new Attribute("xlink:type", "http://www.w3.org/1998/xlink", "sikmple");
        Element test2 = new Element("test");
        try {
           test2.add(a3); 
           test2.add(a4); 
           fail("added conflicting attributes"); 
        }
        catch (Exception e) {
            // success   
        }
 
        
        
    }
    

    public void testCopy() {

        String name = "red:sakjdhjhd";
        String uri = "http://www.red.com/";
        String prefix = "red";
        Element e = new Element(name, uri);

        e.addNamespaceDeclaration("blue", "http://www.blue.com");
        e.addNamespaceDeclaration("green", "http://www.green.com");
        Attribute a1 = new Attribute("test", "test");
        Attribute a2 = new Attribute("pre1:green", "http://www.green.com/", "data");
        Attribute a3 = new Attribute("yellow:sfsdadf", "http://www.yellow.com/", "data");
        e.add(a1);
        e.add(a2);
        e.add(a3);
        
        
        Element e2 = new Element("mauve:child", "http://www.mauve.com");
        e.appendChild(e2);
        
        Element e3 = new Element("mauve:child", "http://www.mauve.com");
        e.insertChild(0, e3);
        Element e4 = new Element("mauve:child", "http://www.mauve.com");
        e3.insertChild(0, e4);

        Element copy = (Element) e.copy();
        
        Assert.assertEquals(e.getNamespaceURI("red"), copy.getNamespaceURI("red"));
        Assert.assertEquals(e.getNamespaceURI("blue"), copy.getNamespaceURI("blue"));
        Assert.assertEquals(e.getValue(), copy.getValue());
        Assert.assertEquals(e.getAttribute("test").getValue(), copy.getAttribute("test").getValue());

    }

}
