/* Copyright © 2002 Elliotte Rusty Harold
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of version 2.1 of 
   the GNU Lesser General Public License as published by the 
   Free Software Foundation.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   
   You can contact Elliotte Rusty Harold by sending e-mail to
   elharo@metalab.unc.edu. Please include the word "XOM" in the
   subject line. The XOM home page is temporarily located at
   http://www.cafeconleche.org/XOM/  but will eventually move
   to http://www.xom.nu/  
*/

package nu.xom.tests;

import junit.framework.TestCase;
import nu.xom.Element;
import nu.xom.Text;
import nu.xom.TreeNode;
import nu.xom.Document;
import nu.xom.IllegalAddException;
import nu.xom.MultipleParentException;
import nu.xom.NoSuchChildException;


/**
 * @author Elliotte Rusty Harold
 *
 */
public class TreeNodeTest extends TestCase {

    public TreeNodeTest(String name) {
        super(name);
    }
    
    Element empty;
    Element notEmpty;
    Text child;
    
    protected void setUp() {
        empty = new Element("Empty");
        notEmpty = new Element("NotEmpty");
        child = new Text("Hello");
        notEmpty.appendChild(child);
    }
    
    public void testHasChildren() {
        
        assertTrue(notEmpty.hasChildren());
        assertTrue(!empty.hasChildren());
        
    } 

    public void testAppendChild() {
        
        Element child = new Element("test");
        empty.appendChild(child);
        assertTrue(empty.hasChildren());
        assertEquals(empty.getFirstChild(), child);
        child.detach();
        
        notEmpty.appendChild(child);
        assertTrue(!notEmpty.getFirstChild().equals(child));
        assertTrue(notEmpty.getFirstChild().getNextSibling().equals(child));
    } 

    public void testAppendChild2() {
        
        try {
            empty.appendChild(new Document(notEmpty));
            fail("appended a document to an element");
        }
        catch (IllegalAddException ex) {
            // success   
        }
        
        try {
            empty.appendChild(child);
            fail("appended a child twice");
        }
        catch (MultipleParentException ex) {
            // success   
        }

    } 

    public void testReplaceChild() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        empty.replaceChild(old1, new1);
        empty.replaceChild(old3, new3);
        empty.replaceChild(old2, new2);

        TreeNode current1 = empty.getFirstChild();
        TreeNode current2 = current1.getNextSibling();
        TreeNode current3 = current2.getNextSibling();
        
        assertEquals(current1, new1);
        assertEquals(current2, new2);
        assertEquals(current3, new3);

    } 

    public void testGetChild() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        assertEquals(old1, empty.getChild(0));
        assertEquals(old3, empty.getChild(2));
        assertEquals(old2, empty.getChild(1));

        try {
            empty.getChild(5);
            fail("No index exception");   
        }
        catch (IndexOutOfBoundsException ex) {
            // success
        }

    } 

   public void testRemoveChild() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        empty.removeChild(1);
        assertEquals(old1, empty.getFirstChild());
        assertEquals(old3, empty.getChild(1));

        try {
            empty.removeChild(5);
            fail("No index exception");   
        }
        catch (IndexOutOfBoundsException ex) {
            // success
        }
        
        empty.removeChild(1);
        empty.removeChild(0);
        assertNull(old2.getParent());
        
        assertEquals(0, empty.getChildCount());

        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        assertEquals(3, empty.getChildCount());

        empty.removeChild(old3);
        empty.removeChild(old1);
        empty.removeChild(old2);
        
        assertEquals(0, empty.getChildCount());
        assertNull(old1.getParent());
        

    } 

    public void testInsertBefore() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        empty.insertBefore(old1, new1);
        empty.insertBefore(old2, new2);
        empty.insertBefore(old3, new3);

        assertEquals(new1, old1.getPreviousSibling());
        assertEquals(new2, old2.getPreviousSibling());
        assertEquals(new3, old3.getPreviousSibling());
        
        Element notInTree = new Element("test");
        try {
           empty.insertBefore(notInTree, new Text("sjhj"));
           fail("Inserted before element not in tree");
        }
        catch (NoSuchChildException ex) {
            // success   
        }
        

    } 

    public void testInsertString() {
        
        empty.insertChild(0, "test");

        String s = null;
        empty.insertChild(0, s);
        // nothing should be inserted
        assertEquals(empty.getChildCount(), 1);
        

    } 

    public void testInsertBeforeFailures() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        try {
            empty.insertBefore(new1, new2);   
        }
        catch (NoSuchChildException ex) {
            // success   
        }
        
        try {
            empty.insertBefore(null, new2);   
        }
        catch (NoSuchChildException ex) {
            // success   
        }
        
        try {
            empty.insertBefore(old2, null);   
        }
        catch (NullPointerException ex) {
            // success   
        }
        

    } 

    public void testInsertAfterFailures() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        try {
            empty.insertAfter(new1, new2);   
        }
        catch (NoSuchChildException ex) {
            // success   
        }
        
        try {
            empty.insertAfter(null, new2);   
        }
        catch (NoSuchChildException ex) {
            // success   
        }
        
        try {
            empty.insertAfter(old2, null);   
        }
        catch (NullPointerException ex) {
            // success   
        }
        

    } 

    public void testInsertAfter() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        empty.appendChild(old3);
        
        empty.insertAfter(old1, new1);
        empty.insertAfter(old2, new2);
        empty.insertAfter(old3, new3);

        assertEquals(new1, old1.getNextSibling());
        assertEquals(new2, old2.getNextSibling());
        assertEquals(new3, old3.getNextSibling());

    } 

   public void testReplaceChildFailures() {
        
        Element old1 = new Element("old1");
        Element old2 = new Element("old2");
        Element old3 = new Element("old3");
        Element new1 = new Element("new1");
        Element new2 = new Element("new2");
        Element new3 = new Element("new3");
        
        empty.appendChild(old1);
        empty.appendChild(old2);
        
        try {
            empty.replaceChild(old3, new3);
            fail("Replaced non-existent child");
        }
        catch (NoSuchChildException ex) {
            // success   
        }

        try {
            empty.replaceChild(old1, null);
        }
        catch (NullPointerException ex) {
            // success   
        }

        try {
            empty.replaceChild(null, new1);
        }
        catch (NoSuchChildException ex) {
            // success   
        }

    } 


}
